"""Configuration files for the Sphinx documentation build for simtools."""

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use Path.resolve() to make it absolute, like shown here.

# pylint: skip-file
from pathlib import Path
from sys import path

import toml
import yaml

import simtools.version

path[:0] = [
    str(Path(p).resolve())
    for p in ["../../src/simtools", "../../src/simtools/applications", "../.."]
]


def get_authors_from_citation_file():
    """Read list of authors from CITATION.cff file."""
    with open(Path(__file__).parent / "../../CITATION.cff", encoding="utf-8") as file:
        citation = yaml.safe_load(file)

    tmp_author = ""
    try:
        for person in citation["authors"]:
            tmp_author = tmp_author + person["given-names"] + " " + person["family-names"]
            tmp_author += " (" + person["affiliation"] + "), "
    except KeyError:
        pass
    return tmp_author[:-2]


def get_python_version_from_pyproject():
    """Read python version from pyproject.toml file."""
    with open(Path(__file__).parent / "../../pyproject.toml") as file:
        pyproject = toml.load(file)

    return (
        pyproject["project"]["requires-python"],
        pyproject["project"]["requires-python"].replace(">", "").replace("=", ""),
    )


# -- Project information -----------------------------------------------------

project = "simtools"
copyright = "2024-2025, gammasim-tools, simtools developers"  # noqa A001
author = get_authors_from_citation_file()
rst_epilog = f"""
.. |author| replace:: {author}
"""

python_min_requires, python_requires = get_python_version_from_pyproject()
rst_epilog = f"""
.. |python_min_requires| replace:: {python_min_requires}
"""

myst_substitutions = {
    "python_min_requires": python_min_requires,
}

# The short X.Y version
version = str(simtools.version.__version__)
# The full version, including alpha/beta/rc tags
release = version

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
    "myst_parser",
    "numpydoc",
    "sphinx_design",
]

# Display todos by setting to True
todo_include_todos = True

# Change the look of autodoc classes
numpydoc_show_class_members = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = [".rst"]

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
# pygments_style = "sphinx"
default_role = "py:obj"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

html_title = f"{project}"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_logo = "../_static/simtools_logo.png"
html_theme_options = {
    "navbar_start": ["navbar-logo"],
    "navbar_end": ["version-switcher", "navbar-icon-links"],
    "show_toc_level": 1,
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/gammasim/simtools",
            "icon": "fa-brands fa-github",
        },
    ],
    "navigation_with_keys": False,
    "switcher": {
        "json_url": "https://raw.githubusercontent.com/gammasim/simtools/refs/heads/main/docs/_static/switcher.json",
        "version_match": "latest",
    },
    "check_switcher": True,
    "logo": {
        "alt_text": "simtools",
    },
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# Output file base name for HTML help builder.
htmlhelp_basename = "simtoolsdoc"

# -- Options for intersphinx extension ---------------------------------------
intersphinx_mapping = {
    "python": (f"https://docs.python.org/{python_requires}", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy", None),
    "astropy": ("https://docs.astropy.org/en/latest", None),
    #    "matplotlib": ("https://matplotlib.org/stable", None),
}

# local dir
locale_dirs = []

# myst (markdown options)
myst_heading_anchors = 3
myst_enable_extensions = [
    "colon_fence",
    "deflist",
    "html_admonition",
    "html_image",
    "substitution",
    "tasklist",
    "attrs_block",
    "attrs_inline",
]


suppress_warnings = ["myst.*", "myst.duplicate_def"]

# Links to be ignored by linkcheck
linkcheck_ignore = [
    r"https://www\.mpi-hd\.mpg\.de/hfm/~bernlohr/sim_telarray",
    r"https://www\.mpi-hd\.mpg\.de/hfm/CTA/MC/Software/Testing/",
    r"https://jama\.cta-observatory\.org/perspective\.req.*",
    r"https://gitlab\.cta-observatory\.org/cta-computing/dpps/simpipe/corsika-opt-patches",
    r"https://gitlab\.cta-observatory\.org/Konrad\.Bernloehr/sim_telarray",
]



"""Generate run scripts and directories for CORSIKA simulations."""

import logging
import stat
from pathlib import Path

from simtools.io import io_handler
from simtools.runners.runner_services import RunnerServices


class MissingRequiredEntryInCorsikaConfigError(Exception):
    """Exception for missing required entry in corsika config."""


class CorsikaRunner:
    """
    Generate run scripts and directories for CORSIKA simulations. Run simulations if requested.

    CorsikaRunner is responsible for configuring and running CORSIKA, using corsika_autoinputs
    provided by the sim_telarray package. CorsikaRunner generates shell scripts to be run
    externally or by the simulator module simulator.

    CorsikaRunner is configured through a CorsikaConfig instance.

    Parameters
    ----------
    corsika_config_data: CorsikaConfig
        CORSIKA configuration.
    simtel_path: str or Path
        Location of source of the sim_telarray/CORSIKA package.
    label: str
        Instance label.
    keep_seeds: bool
        Use seeds based on run number and primary particle. If False, use sim_telarray seeds.
    use_multipipe: bool
        Use multipipe to run CORSIKA and sim_telarray.
    """

    def __init__(
        self,
        corsika_config,
        simtel_path,
        label=None,
        keep_seeds=False,
        use_multipipe=False,
    ):
        """Initialize CorsikaRunner."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init CorsikaRunner")
        self.label = label

        self.corsika_config = corsika_config
        self._keep_seeds = keep_seeds
        self._use_multipipe = use_multipipe

        self._simtel_path = Path(simtel_path)
        self.io_handler = io_handler.IOHandler()

        self.runner_service = RunnerServices(corsika_config, label)
        self._directory = self.runner_service.load_data_directories("corsika")

    def prepare_run_script(
        self, run_number=None, extra_commands=None, input_file=None, use_pfp=True
    ):
        """
        Get the full path of the run script file for a given run number.

        Parameters
        ----------
        use_pfp: bool
            Whether to use the preprocessor in preparing the CORSIKA input file
        run_number: int
            Run number.
        extra_commands: str
            Additional commands for running simulations.

        Returns
        -------
        Path:
            Full path of the run script file.
        """
        if input_file is not None:
            self._logger.warning(
                "input_file parameter is not used in CorsikaRunner.prepare_run_script"
            )
        self.corsika_config.run_number = run_number

        script_file_path = self.get_file_name(
            file_type="sub_script", run_number=self.corsika_config.run_number
        )
        corsika_input_file = self.corsika_config.generate_corsika_input_file(
            use_multipipe=self._use_multipipe, use_test_seeds=self._keep_seeds
        )

        # CORSIKA input file for a specific run, created by the preprocessor pfp
        corsika_input_tmp_name = self.corsika_config.get_corsika_config_file_name(
            file_type="config_tmp", run_number=self.corsika_config.run_number
        )
        corsika_input_tmp_file = self._directory["inputs"].joinpath(corsika_input_tmp_name)
        # CORSIKA log file naming (temporary and final)
        corsika_log_tmp_file = (
            self._directory["data"]
            .joinpath(f"run{self.corsika_config.run_number:06}")
            .joinpath(f"run{self.corsika_config.run_number}.log")
        )
        corsika_log_file = self.get_file_name(
            file_type="corsika_log", run_number=self.corsika_config.run_number
        )

        if use_pfp:
            pfp_command = self._get_pfp_command(corsika_input_tmp_file, corsika_input_file)
        autoinputs_command = self._get_autoinputs_command(
            self.corsika_config.run_number, corsika_input_tmp_file
        )

        self._logger.debug(f"Extra commands to be added to the run script: {extra_commands}")
        self._logger.debug(f"CORSIKA data will be set to {self._directory['data']}")

        with open(script_file_path, "w", encoding="utf-8") as file:
            file.write("#!/usr/bin/env bash\n")
            file.write("set -e\n")
            file.write("set -o pipefail\n")

            # Setting SECONDS variable to measure runtime
            file.write("\nSECONDS=0\n")

            if extra_commands is not None:
                file.write("\n# Writing extras\n")
                file.write(f"{extra_commands}\n")
                file.write("# End of extras\n\n")

            file.write(f"export CORSIKA_DATA={self._directory['data']}\n")
            file.write('mkdir -p "$CORSIKA_DATA"\n')
            file.write('cd "$CORSIKA_DATA" || exit 2\n')
            if use_pfp:
                file.write("\n# Running pfp\n")
                file.write(pfp_command)
                file.write("\n# Replacing the XXXXXX placeholder with the run number\n")
                file.write(
                    f"sed -i 's/XXXXXX/{self.corsika_config.run_number:06}/g' "
                    f"{corsika_input_tmp_file}\n"
                )
            else:
                file.write("\n# Copying CORSIKA input file to run location\n")
                file.write(f"cp {corsika_input_file} {corsika_input_tmp_file}")
            file.write("\n# Running corsika_autoinputs\n")
            file.write(autoinputs_command)
            file.write("\n# Moving log files to the corsika log directory\n")
            file.write(f"gzip {corsika_log_tmp_file}\n")
            file.write(f"mv -v {corsika_log_tmp_file}.gz {corsika_log_file}\n")

            file.write('\necho "RUNTIME: $SECONDS"\n')

        script_file_path.chmod(script_file_path.stat().st_mode | stat.S_IXUSR | stat.S_IXGRP)
        return script_file_path

    def get_resources(self, run_number=None):
        """Return computing resources used."""
        return self.runner_service.get_resources(run_number)

    def _get_pfp_command(self, input_tmp_file, corsika_input_file):
        """
        Get pfp pre-processor command.

        pfp is a pre-processor tool and part of sim_telarray.

        Parameters
        ----------
        input_tmp_file: Path
            Temporary input file.

        Returns
        -------
        str
            pfp command.
        """
        cmd = self._simtel_path.joinpath("sim_telarray/bin/pfp")
        cmd = str(cmd) + f" -V -DWITHOUT_MULTIPIPE - < {corsika_input_file}"
        cmd += f" > {input_tmp_file} || exit\n"
        return cmd

    def _get_autoinputs_command(self, run_number, input_tmp_file):
        """
        Get autoinputs command.

        corsika_autoinputs is a tool to generate random and user/host dependent
        parameters for CORSIKA configuration.

        Parameters
        ----------
        run_number: int
            Run number.
        input_tmp_file: Path
            Temporary input file.

        Returns
        -------
        str
            autoinputs command.
        """
        corsika_bin_path = self._simtel_path.joinpath("corsika-run/corsika")

        log_file = self.get_file_name(file_type="log", run_number=run_number)
        if self._use_multipipe:
            log_file = log_file.with_name(f"multipipe_{log_file.name}")

        cmd = self._simtel_path.joinpath("sim_telarray/bin/corsika_autoinputs")
        cmd = str(cmd) + f" --run {corsika_bin_path}"
        cmd += f" -R {run_number}"
        cmd += ' -p "$CORSIKA_DATA"'
        if self._keep_seeds:
            logging.warning(
                "Using --keep-seeds option in corsika_autoinputs is not recommended. "
                "It should only be used for testing purposes."
            )
            cmd += " --keep-seeds"
        cmd += f" {input_tmp_file} | gzip > {log_file} 2>&1"
        cmd += " || exit 1\n"
        return cmd

    def get_file_name(
        self,
        simulation_software="corsika",
        file_type=None,
        run_number=None,
        mode="",
        model_version_index=0,
    ):
        """
        Get the full path of a file for a given run number.

        Parameters
        ----------
        simulation_software: str
            Simulation software.
        file_type: str
            File type.
        run_number: int
            Run number.
        model_version_index: int
            Index of the model version.
            This is used to select the correct simulator_array instance in case
            multiple array models are simulated.

        Returns
        -------
        str
            File name with full path.
        """
        if simulation_software.lower() != "corsika":
            raise ValueError(
                f"simulation_software ({simulation_software}) is not supported in CorsikaRunner"
            )
        return self.runner_service.get_file_name(
            file_type=file_type,
            run_number=run_number,
            mode=mode,
            _model_version_index=model_version_index,
        )



"""Definition of the ArrayModel class."""

import logging
from pathlib import Path

import astropy.units as u
from astropy.table import QTable

from simtools.data_model import data_reader, schema
from simtools.io import io_handler
from simtools.model.calibration_model import CalibrationModel
from simtools.model.site_model import SiteModel
from simtools.model.telescope_model import TelescopeModel
from simtools.simtel.simtel_config_writer import SimtelConfigWriter
from simtools.utils import general, names


class ArrayModel:
    """
    Representation of an observatory consisting of site, telescopes, and further devices.

    Parameters
    ----------
    db_config: dict
        Database configuration.
    model_version: str
        Model version.
    label: str, optional
        Instance label. Used for output file naming.
    site: str, optional
        Site name.
    layout_name: str, optional
        Layout name.
    array_elements: Union[str, Path, List[str]], optional
        Array element definitions (list of array element or path to file with
        the array element positions).
    calibration_device_types: List[str], optional
        List of calibration device types (e.g., 'flat_fielding') attached to each telescope.
    sim_telarray_seeds : dict, optional
        Dictionary with configuration for sim_telarray random instrument setup.
    simtel_path: str, Path, optional
        Path to the sim_telarray installation directory.
    overwrite_model_parameters: str, optional
        File name to overwrite model parameters from DB with provided values.
    """

    def __init__(
        self,
        db_config,
        model_version,
        label=None,
        site=None,
        layout_name=None,
        array_elements=None,
        calibration_device_types=None,
        sim_telarray_seeds=None,
        simtel_path=None,
        overwrite_model_parameters=None,
    ):
        """Initialize ArrayModel."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init ArrayModel")
        self.db_config = db_config
        self.model_version = model_version
        self.label = label
        self.layout_name = (
            layout_name[0]
            if isinstance(layout_name, list) and len(layout_name) == 1
            else layout_name
        )
        self._config_file_path = None
        self._config_file_directory = None
        self.io_handler = io_handler.IOHandler()

        self.overwrite_model_parameters = overwrite_model_parameters

        self.array_elements, self.site_model, self.telescope_models, self.calibration_models = (
            self._initialize(site, array_elements, calibration_device_types)
        )

        self._telescope_model_files_exported = False
        self._array_model_file_exported = False
        self.sim_telarray_seeds = sim_telarray_seeds
        self.simtel_path = simtel_path

    def _initialize(self, site, array_elements_config, calibration_device_types):
        """
        Initialize ArrayModel taking different configuration options into account.

        Parameters
        ----------
        site: str
            Site name.
        array_elements_config: Union[str, Path, List[str]]
            Array element definitions.
        calibration_device_types: str
            Calibration device types.

        Returns
        -------
        dict
            Dict with telescope positions.
        SiteModel
            Site model.
        dict
            Dict with telescope models.
        """
        self._logger.debug(f"Getting site parameters from DB ({site})")
        site_model = SiteModel(
            site=names.validate_site_name(site),
            db_config=self.db_config,
            model_version=self.model_version,
            label=self.label,
            overwrite_model_parameters=self.overwrite_model_parameters,
        )

        # Case 1: array_elements is a file name
        if isinstance(array_elements_config, str | Path):
            array_elements = self._load_array_element_positions_from_file(
                array_elements_config, site
            )
        # Case 2: array elements is a list of elements
        elif isinstance(array_elements_config, list):
            array_elements = self._get_array_elements_from_list(array_elements_config, site_model)
        # Case 3: array elements defined in DB by array layout name
        elif self.layout_name is not None:
            array_elements = self._get_array_elements_from_list(
                site_model.get_array_elements_for_layout(self.layout_name)
            )
        else:
            raise ValueError(
                "No array elements found. "
                "Possibly missing valid layout name or missing telescope list."
            )

        telescope_models, calibration_models = self._build_telescope_models(
            site_model, array_elements, calibration_device_types
        )

        return array_elements, site_model, telescope_models, calibration_models

    @property
    def config_file_path(self):
        """
        Return the path of the array config file for sim_telarray.

        Returns
        -------
        Path
            Path of the exported config file for sim_telarray.
        """
        if self._config_file_path is None:
            config_file_name = names.simtel_config_file_name(
                array_name=self.layout_name,
                site=self.site_model.site,
                model_version=self.model_version,
                label=self.label,
            )
            self._config_file_path = self.get_config_directory().joinpath(config_file_name)
        return self._config_file_path

    @property
    def number_of_telescopes(self):
        """
        Return the number of telescopes.

        Returns
        -------
        int
            Number of telescopes.
        """
        return len(self.telescope_models)

    @property
    def site(self):
        """
        Return site.

        Returns
        -------
        str
            Site name.
        """
        return self.site_model.site

    def _build_telescope_models(self, site_model, array_elements, calibration_device_types):
        """
        Build telescope models for all telescopes of this array.

        Adds calibration device models, if requested through the calibration_device_types argument.
        Calibration device models are stored in a dictionary with the telescope name as key (to
        identify the calibration device model on a given telescope).

        Includes reading of telescope model parameters from the database.
        The array is defined in the array_elements dictionary. Array element positions
        are read from the database if no values are given in this dictionary.

        Parameters
        ----------
        site_model: SiteModel
            Site model.
        array_elements: dict
            Dict with array elements.
        calibration_device_types: List[str]
            List of calibration device types (e.g., 'flat_fielding')

        Returns
        -------
        dict, dict
            Dictionaries with telescope and calibration models.
        """
        telescope_models, calibration_models = {}, {}

        for element_name in array_elements:
            if names.get_collection_name_from_array_element_name(element_name) != "telescopes":
                continue

            telescope_models[element_name] = TelescopeModel(
                site=site_model.site,
                telescope_name=element_name,
                model_version=self.model_version,
                db_config=self.db_config,
                label=self.label,
                overwrite_model_parameters=self.overwrite_model_parameters,
            )
            calibration_models[element_name] = self._build_calibration_models(
                telescope_models[element_name],
                site_model,
                calibration_device_types,
            )

        return telescope_models, calibration_models

    def _build_calibration_models(self, telescope_model, site_model, calibration_device_types):
        """
        Build calibration device models for all telescopes in the array.

        A telescope can have multiple calibration devices of different types.

        Returns
        -------
        dict
            Dict with calibration device models.
        """
        calibration_models = {}
        for calibration_device_type in calibration_device_types or []:
            device_name = telescope_model.get_calibration_device_name(calibration_device_type)
            if device_name is None:
                continue

            calibration_models[device_name] = CalibrationModel(
                site=site_model.site,
                calibration_device_model_name=device_name,
                db_config=self.db_config,
                model_version=self.model_version,
                label=self.label,
                overwrite_model_parameters=self.overwrite_model_parameters,
            )
        return calibration_models

    def print_telescope_list(self):
        """Print list of telescopes."""
        for tel_name, data in self.telescope_models.items():
            print(f"Name: {tel_name}\t Model: {data.name}")

    def export_simtel_telescope_config_files(self):
        """Export sim_telarray configuration files for all telescopes into the model directory."""
        exported_models = []
        for tel_model in self.telescope_models.values():
            name = tel_model.name
            if name not in exported_models:
                self._logger.debug(f"Exporting configuration file for telescope {name}")
                tel_model.write_sim_telarray_config_file(
                    additional_models=self.calibration_models.get(tel_model.name)
                )
                exported_models.append(name)
            else:
                self._logger.debug(
                    f"Configuration file for telescope {name} already exists - skipping"
                )

        self._telescope_model_files_exported = True

    def export_sim_telarray_config_file(self):
        """Export sim_telarray configuration file for the array into the model directory."""
        self.site_model.export_model_files()

        self._logger.info(f"Writing array configuration file into {self.config_file_path}")
        simtel_writer = SimtelConfigWriter(
            site=self.site_model.site,
            layout_name=self.layout_name,
            model_version=self.model_version,
            label=self.label,
            simtel_path=self.simtel_path,
        )
        simtel_writer.write_array_config_file(
            config_file_path=self.config_file_path,
            telescope_model=self.telescope_models,
            site_model=self.site_model,
            additional_metadata=self._get_additional_simtel_metadata(),
        )
        self._array_model_file_exported = True

    def export_all_simtel_config_files(self):
        """
        Export sim_telarray config file for the array and for each individual telescope.

        Config files are exported into the output model directory.
        """
        if not self._telescope_model_files_exported:
            self.export_simtel_telescope_config_files()
        if not self._array_model_file_exported:
            self.export_sim_telarray_config_file()

    def get_config_directory(self):
        """
        Get the path of the array config directory for sim_telarray.

        Returns
        -------
        Path
            Path of the config directory path for sim_telarray.
        """
        if self._config_file_directory is None:
            self._config_file_directory = self.io_handler.get_model_configuration_directory(
                model_version=self.model_version
            )
        return self._config_file_directory

    def pack_model_files(self):
        """
        Pack all model files into a tar.gz archive.

        Returns
        -------
        Path
            Path of the packed model files archive.
        """
        model_files = list(Path(self.get_config_directory()).rglob("*"))
        if not model_files:
            self._logger.warning("No model files found to pack.")
            return None

        archive_name = self.get_config_directory() / f"model_files_{self.model_version}.tar.gz"
        general.pack_tar_file(archive_name, model_files, sub_dir=f"model/{self.model_version}")
        self._logger.info(f"Packed model files into {archive_name}")
        return archive_name

    def _load_array_element_positions_from_file(self, array_elements_file, site):
        """
        Load array element (e.g. telescope) positions from a file into a dict.

        Dictionary format: {telescope_name: {position_x: x, position_y: y, position_z: z}}

        Parameters
        ----------
        array_elements_file: Union[str, Path]
            Path to the file with the array element positions.
        site: str
            Site name.

        Returns
        -------
        dict
            Dict with telescope positions.
        """
        table = data_reader.read_table_from_file(file_name=array_elements_file)

        return {
            row["telescope_name"]: self._get_telescope_position_parameter(
                row["telescope_name"], site, row["position_x"], row["position_y"], row["position_z"]
            )
            for row in table
        }

    def _get_telescope_position_parameter(
        self, telescope_name, site, x, y, z, parameter_version=None
    ):
        """
        Return dictionary with telescope position parameters (following DB model database format).

        Parameters
        ----------
        telescope_name: str
            Name of the telescope.
        site: str
            Site name.
        x: astropy.Quantity
            X ground position.
        y: astropy.Quantity
            Y ground position.
        z: astropy.Quantity
            Z ground position.

        Returns
        -------
        dict
            Dict with telescope position parameters.
        """
        return {
            "schema_version": schema.get_model_parameter_schema_version(),
            "parameter": "array_element_position_ground",
            "instrument": telescope_name,
            "site": site,
            "parameter_version": parameter_version,
            "unique_id": None,
            "value": [x.to("m").value, y.to("m").value, z.to("m").value],
            "unit": "m",
            "type": "float64",
            "file": False,
            "meta_parameter": False,
            "model_parameter_schema_version": "0.1.0",
        }

    def _get_array_elements_from_list(self, array_elements_list, site_model=None):
        """
        Return dictionary with array elements from a list of telescope names.

        Input list can contain telescope names (e.g, LSTN-01) or a telescope
        type (e.g., MSTN). In the latter case, all telescopes of this specific
        type are added.

        Parameters
        ----------
        array_elements_list: list
            List of telescope names.
        site_model: SiteModel
            Site model.

        Returns
        -------
        dict
            Dict with array elements.
        """
        array_elements_dict = {}
        for name in array_elements_list:
            try:
                array_elements_dict[names.validate_array_element_name(name)] = None
            except ValueError:
                array_elements_dict.update(self._get_all_array_elements_of_type(name, site_model))
        return array_elements_dict

    def _get_all_array_elements_of_type(self, array_element_type, site_model):
        """
        Return all array elements of a specific type.

        Parameters
        ----------
        array_element_type : str
            Type of the array element (e.g. LSTN, MSTS)
        site_model: SiteModel
            Site model.

        Returns
        -------
        dict
            Dict with array elements.
        """
        return self._get_array_elements_from_list(
            site_model.get_array_elements_of_type(array_element_type)
        )

    def export_array_elements_as_table(self, coordinate_system="ground"):
        """
        Export array elements positions to astropy table.

        Parameters
        ----------
        coordinate_system: str
            Positions are exported in this coordinate system.

        Returns
        -------
        astropy.table.QTable
            Astropy table with the telescope layout information.
        """
        table = QTable(meta={"array_name": self.layout_name, "site": self.site_model.site})

        name, pos_x, pos_y, pos_z, tel_r = [], [], [], [], []
        for tel_name, data in self.telescope_models.items():
            name.append(tel_name)
            xyz = data.position(coordinate_system=coordinate_system)
            pos_x.append(xyz[0])
            pos_y.append(xyz[1])
            pos_z.append(xyz[2])
            try:
                # add tests of KeyError after positions calibration_elements are added to DB
                tel_r.append(data.get_parameter_value_with_unit("telescope_sphere_radius"))
            except KeyError:  # not all array elements have a sphere radius
                tel_r.append(0.0 * u.m)

        table["telescope_name"] = name
        if coordinate_system == "ground":
            table["position_x"] = pos_x
            table["position_y"] = pos_y
            table["position_z"] = pos_z
        elif coordinate_system == "utm":
            table["utm_east"] = pos_x
            table["utm_north"] = pos_y
            table["altitude"] = pos_z
        table["sphere_radius"] = tel_r

        table.sort("telescope_name")
        return table

    def _get_additional_simtel_metadata(self):
        """
        Collect additional metadata to be included in sim_telarray output.

        Returns
        -------
        dict
            Dictionary with additional metadata.
        """
        metadata = {}
        if self.sim_telarray_seeds is not None:
            metadata.update(self.sim_telarray_seeds)

        metadata["nsb_integrated_flux"] = self.site_model.get_nsb_integrated_flux()

        return metadata



"""Prepare layout for coordinate transformations."""

import json
import logging
from pathlib import Path

import astropy.units as u
import numpy as np
from astropy.table import QTable

import simtools.utils.general as gen
from simtools.data_model import data_reader, schema
from simtools.io import io_handler
from simtools.layout.geo_coordinates import GeoCoordinates
from simtools.layout.telescope_position import TelescopePosition
from simtools.model.site_model import SiteModel
from simtools.model.telescope_model import TelescopeModel
from simtools.utils import names, value_conversion


class InvalidTelescopeListFileError(Exception):
    """Exception for invalid telescope list file."""


class InvalidCoordinateDataTypeError(Exception):
    """Exception for low-precision coordinate data type."""


class ArrayLayout:
    """
    Manage telescope positions at the array layout level.

    Parameters
    ----------
    db_config: dict
        Database configuration.
    site: str
        Site name or location (e.g., North/South or LaPalma/Paranal)
    model_version: str
        Version of the model (e.g., 6.0.0).
    label: str
        Instance label.
    name: str
        Name of the layout.
    telescope_list_file: str or Path
        Path to the telescope list file.
    telescope_list_metadata_file: str or Path
        Path to telescope list metadata (if not part of telescope_list_file)
    validate: bool
        Validate input file list.
    """

    def __init__(
        self,
        db_config,
        site,
        model_version,
        label=None,
        name=None,
        telescope_list_file=None,
        telescope_list_metadata_file=None,
        validate=False,
    ):
        """Initialize ArrayLayout."""
        self._logger = logging.getLogger(__name__)

        self.model_version = model_version
        self.label = label
        self.name = name
        self.db_config = db_config
        self.site = None if site is None else names.validate_site_name(site)
        self.site_model = None
        self.io_handler = io_handler.IOHandler()
        self.geo_coordinates = GeoCoordinates()

        self._telescope_list = []
        self._corsika_observation_level = None
        self._reference_position_dict = {}
        self._array_center = None

        self._initialize_array_layout(
            telescope_list_file=telescope_list_file,
            telescope_list_metadata_file=telescope_list_metadata_file,
            validate=validate,
        )

    def __len__(self):
        """Return number of telescopes in the layout."""
        return len(self._telescope_list)

    def __getitem__(self, i):
        """Return telescope at list position i."""
        return self._telescope_list[i]

    def _initialize_site_parameters_from_db(self):
        """Initialize site parameters required for transformations using the database."""
        self._logger.debug("Initialize parameters from DB")
        if self.db_config is None:
            raise ValueError("No database configuration provided")

        self.site_model = SiteModel(
            site=self.site,
            model_version=self.model_version,
            db_config=self.db_config,
        )
        self._corsika_observation_level = self.site_model.get_corsika_site_parameters().get(
            "corsika_observation_level", None
        )
        self._reference_position_dict = self.site_model.get_reference_point()
        self._logger.debug(f"Reference point: {self._reference_position_dict}")

    def _initialize_coordinate_systems(self):
        """
        Initialize array center and coordinate systems.

        By definition, the array center is at (0,0) in
        the ground coordinate system.

        Raises
        ------
        TypeError
            invalid array center definition.

        """
        self._array_center = TelescopePosition()
        self._array_center.name = "array_center"
        self._array_center.set_coordinates("ground", 0.0 * u.m, 0.0 * u.m)
        self._set_array_center_utm()
        self._array_center.set_altitude(
            u.Quantity(self._reference_position_dict.get("center_altitude", np.nan * u.m))
        )
        _name = self._reference_position_dict.get("array_name")
        self.name = _name if _name is not None else self.name

        self._logger.debug(f"Initialized array center at UTM {self._reference_position_dict}")
        self._array_center.convert_all(
            crs_local=self.geo_coordinates.crs_local(self._array_center),
            crs_wgs84=self.geo_coordinates.crs_wgs84(),
            crs_utm=self.geo_coordinates.crs_utm(
                self._reference_position_dict.get("epsg_code", None)
            ),
        )

    def _set_array_center_utm(self):
        """
        Set array center coordinates in UTM system.

        Convert array center position to WGS84 system
        (as latitudes are required for the definition
        for the definition of the ground coordinate system)

        """
        self._array_center.set_coordinates(
            "utm",
            u.Quantity(self._reference_position_dict.get("center_easting", np.nan * u.m)),
            u.Quantity(self._reference_position_dict.get("center_northing", np.nan * u.m)),
        )
        self._array_center.convert_all(
            crs_local=None,
            crs_wgs84=self.geo_coordinates.crs_wgs84(),
            crs_utm=self.geo_coordinates.crs_utm(
                self._reference_position_dict.get("epsg_code", None)
            ),
        )

    def _altitude_from_corsika_z(self, pos_z=None, altitude=None, telescope_axis_height=None):
        """
        Calculate altitude.

        The value is calculated from CORSIKA z-coordinate (if pos_z is given) or CORSIKA
        z-coordinate from altitude (if altitude is given).

        Parameters
        ----------
        pos_z: astropy.Quantity
            CORSIKA z-coordinate of telescope in equivalent units of meter.
        altitude: astropy.Quantity
            Telescope altitude in equivalent units of meter.
        tel_axis_height: astropy.Quantity
            Telescope axis height in equivalent units of meter.

        Returns
        -------
        astropy.Quantity
            Altitude or CORSIKA z-coordinate (np.nan in case of ill-defined value).

        """
        self._logger.debug(
            f"pos_z: {pos_z}, altitude: {altitude}, "
            f"axis_height: {telescope_axis_height}, "
            f"obs_level: {self._corsika_observation_level}"
        )

        if pos_z is not None and altitude is None:
            return TelescopePosition.convert_telescope_altitude_from_corsika_system(
                pos_z,
                self._corsika_observation_level,
                telescope_axis_height,
            )

        if altitude is not None and pos_z is None:
            return TelescopePosition.convert_telescope_altitude_to_corsika_system(
                altitude,
                self._corsika_observation_level,
                telescope_axis_height,
            )
        return np.nan

    def _load_telescope_names(self, row):
        """
        Read and set telescope names.

        Parameters
        ----------
        row: astropy table row
            table row.

        Returns
        -------
        tel: TelescopePosition
            Instance of TelescopePosition.

        Raises
        ------
        InvalidTelescopeListFileError
            in case neither telescope name or asset_code / sequence number are given.

        """
        tel = TelescopePosition()
        try:
            tel.name = row["telescope_name"]
            if "asset_code" not in row:
                try:
                    tel.asset_code = names.get_array_element_type_from_name(tel.name)
                # asset code is not a valid telescope name; possibly a calibration device
                except ValueError:
                    tel.asset_code = tel.name.split("-")[0]
        except KeyError:
            pass
        try:
            if tel.name is None:
                tel.name = row["asset_code"] + "-" + row["sequence_number"]
            tel.asset_code = row["asset_code"]
            tel.sequence_number = row["sequence_number"]
        except KeyError:
            pass
        try:
            tel.geo_code = row["geo_code"]
        except KeyError:
            pass
        if tel.name is None:
            msg = "Missing required row with telescope_name or asset_code/sequence_number"
            self._logger.error(msg)
            raise InvalidTelescopeListFileError(msg)

        return tel

    def _try_set_coordinate(self, row, tel, table, crs_name, key1, key2):
        """
        Try and set coordinates for all coordinate systems.

        Parameters
        ----------
        row: dict
            A row of the astropy.table.Table with array element coordinates.
        tel: TelescopePosition
            Instance of TelescopePosition.
        table: astropy.table.Table or astropy.table.QTable
            data table with array element coordinates.
        crs_name: str
            Name of coordinate system.
        key1: str
            Name of x-coordinate.
        key2: str
            Name of y-coordinate.
        """
        try:
            tel.set_coordinates(
                crs_name,
                value_conversion.get_value_as_quantity(row[key1], table[key1].unit),
                value_conversion.get_value_as_quantity(row[key2], table[key2].unit),
            )
        except KeyError:
            pass

    def _try_set_altitude(self, row, tel, table):
        """
        Try and set altitude.

        It sets the altitude of the TelescopePosition instance.

        Parameters
        ----------
        row: dict
            A row of the astropy.table.Table with array element coordinates.
        tel: TelescopePosition
            Instance of TelescopePosition.
        table: astropy.table.Table or astropy.table.QTable
            data table with array element coordinates.
        """
        try:
            tel.set_altitude(
                self._altitude_from_corsika_z(
                    pos_z=value_conversion.get_value_as_quantity(
                        row["position_z"], table["position_z"].unit
                    ),
                    telescope_axis_height=tel.get_axis_height(),
                )
            )
        except KeyError:
            pass
        try:
            tel.set_altitude(
                value_conversion.get_value_as_quantity(row["altitude"], table["altitude"].unit)
            )
        except KeyError:
            pass

    def _initialize_array_layout(
        self, telescope_list_file, telescope_list_metadata_file=None, validate=False
    ):
        """
        Initialize the Layout array including site and telescope parameters.

        Read array list if telescope_list_file is given.

        Parameters
        ----------
        telescope_list_file: str or Path
            Path to the telescope list file.
        telescope_list_metadata_file: str or Path
            Path to the telescope list metadata file.
        validate: bool
            Validate telescope list file against schema.

        Returns
        -------
        astropy.table.QTable
            Table with the telescope layout information.
        """
        self._logger.debug("Initializing array (site and telescope parameters)")
        self._initialize_site_parameters_from_db()
        self._initialize_coordinate_systems()

        if telescope_list_file is None:
            return None

        self._logger.debug(f"Reading telescope list from {telescope_list_file}")
        if Path(telescope_list_file).suffix == ".json":
            table = self._read_table_from_json_file(file_name=telescope_list_file)
        else:
            table = data_reader.read_table_from_file(
                file_name=telescope_list_file,
                validate=validate,
                metadata_file=telescope_list_metadata_file,
            )

        for row in table:
            tel = self._load_telescope_names(row)
            if names.get_collection_name_from_array_element_name(tel.name) == "telescopes":
                self._set_telescope_auxiliary_parameters(tel)
            self._try_set_coordinate(row, tel, table, "ground", "position_x", "position_y")
            self._try_set_coordinate(row, tel, table, "utm", "utm_east", "utm_north")
            self._try_set_coordinate(row, tel, table, "mercator", "latitude", "longitude")
            self._try_set_altitude(row, tel, table)
            self._telescope_list.append(tel)

        return table

    def _read_table_from_json_file(self, file_name):
        """
        Read a telescope position from a json file and return as astropy table.

        Parameters
        ----------
        file_name: str or Path
            Path to the json file.

        Returns
        -------
        astropy.table.QTable
            Table with the telescope layout information.
        """
        with Path(file_name).open("r", encoding="utf-8") as file:
            data = json.load(file)

        position = data["value"]
        if isinstance(position, str):
            position = gen.convert_string_to_list(position)
        self.site = data.get("site", None)

        table = QTable()
        table["telescope_name"] = [data["instrument"]]
        if "utm" in data["parameter"]:
            table["utm_east"] = [position[0]] * u.Unit(data["unit"])
            table["utm_north"] = [position[1]] * u.Unit(data["unit"])
            table["altitude"] = [position[2]] * u.Unit(data["unit"])
        else:
            table["position_x"] = [position[0]] * u.Unit(data["unit"])
            table["position_y"] = [position[1]] * u.Unit(data["unit"])
            table["position_z"] = [position[2]] * u.Unit(data["unit"])
        return table

    def _get_telescope_model(self, telescope_name):
        """
        Get telescope model from the database.

        Parameters
        ----------
        telescope_name: str
            Name of the telescope.

        Returns
        -------
        TelescopeModel
            Telescope model instance.
        """
        return TelescopeModel(
            site=self.site,
            telescope_name=telescope_name,
            model_version=self.model_version,
            db_config=self.db_config,
            label=self.label,
        )

    def _set_telescope_auxiliary_parameters(self, telescope, telescope_name=None):
        """
        Set auxiliary CORSIKA parameters for a telescope.

        Uses as default the design model if telescope is not found in the database.

        Parameters
        ----------
        telescope: TelescopePosition
            Instance of TelescopePosition.

        """
        telescope_name = telescope_name if telescope_name is not None else telescope.name
        if names.get_collection_name_from_array_element_name(telescope_name) == "telescopes":
            self._logger.debug(
                f"Reading auxiliary telescope parameters for {telescope_name}"
                f" (model version {self.model_version})"
            )
            try:
                tel_model = self._get_telescope_model(telescope_name)
            except ValueError:  # telescope not found in the database revert to design model
                tel_model = self._get_telescope_model(
                    names.array_element_design_types(
                        names.get_array_element_type_from_name(telescope_name)
                    )[0]
                )

            for para in ("telescope_axis_height", "telescope_sphere_radius"):
                telescope.set_auxiliary_parameter(
                    para, tel_model.get_parameter_value_with_unit(para)
                )

    def add_telescope(
        self, telescope_name, crs_name, xx, yy, altitude=None, tel_corsika_z=None, design_model=None
    ):
        """
        Add an individual telescope to the telescope list.

        Parameters
        ----------
        telescope_name: str
            Name of the telescope starting with L, M or S (e.g. LST-01, MST-06 ...)
        crs_name: str
            Name of coordinate system
        xx: astropy.units.quantity.Quantity
            x-coordinate for the given coordinate system
        yy: astropy.units.quantity.Quantity
            y-coordinate for the given coordinate system
        altitude: astropy.units.quantity.Quantity
            Altitude coordinate in equivalent units of u.m.
        tel_corsika_z: astropy.units.quantity.Quantity
            CORSIKA z-position (requires setting of CORSIKA observation level and telescope sphere\
            center).
        design_model: str
            Name of the design model (optional).
            If none, telescope type + "-design" is used.
        """
        tel = TelescopePosition(name=telescope_name)
        self._set_telescope_auxiliary_parameters(tel, design_model)
        tel.set_coordinates(crs_name, xx, yy)
        if altitude is not None:
            tel.set_altitude(altitude)
        elif tel_corsika_z is not None:
            tel.set_altitude(
                self._altitude_from_corsika_z(
                    pos_z=tel_corsika_z, telescope_axis_height=tel.get_axis_height()
                )
            )
        self._telescope_list.append(tel)

    def _get_export_metadata(self):
        """
        File metadata for export of array element list to file.

        Included array center definition, CORSIKA telescope parameters, and EPSG code.

        Returns
        -------
        dict
            Metadata header for array element list export.

        """
        _meta = {}
        if self._array_center:
            (
                _meta["center_easting"],
                _meta["center_northing"],
                _meta["center_altitude"],
            ) = self._array_center.get_coordinates("utm")
        _meta["epsg_code"] = self._reference_position_dict.get("epsg_code", None)
        _meta["array_name"] = self.name

        return _meta

    def export_telescope_list_table(self, crs_name):
        """
        Export array elements positions to astropy table.

        Parameters
        ----------
        crs_name: str
            Name of coordinate system to be used for export.

        Returns
        -------
        astropy.table.QTable
            Astropy table with the telescope layout information.
        """
        table = QTable(meta=self._get_export_metadata())

        tel_names, asset_code, sequence_number, geo_code = [], [], [], []
        pos_x, pos_y, pos_z, pos_t, tel_r = [], [], [], [], []
        for tel in self._telescope_list:
            tel_names.append(tel.name)
            asset_code.append(tel.asset_code)
            sequence_number.append(tel.sequence_number)
            geo_code.append(tel.geo_code)
            x, y, z = tel.get_coordinates(crs_name)
            if crs_name == "ground":
                z = self._altitude_from_corsika_z(
                    altitude=z, telescope_axis_height=tel.get_axis_height()
                )
                pos_t.append(tel.get_axis_height())
            pos_x.append(x)
            pos_y.append(y)
            pos_z.append(z)
            tel_r.append(tel.get_sphere_radius())

        # prefer asset_code / sequence_number of telescope_name
        if all(v is not None for v in asset_code) and all(v is not None for v in sequence_number):
            table["asset_code"] = asset_code
            table["sequence_number"] = sequence_number
        else:
            table["telescope_name"] = tel_names
        if any(v is not None for v in geo_code):
            table["geo_code"] = geo_code

        try:
            _name_x, _name_y, _name_z = self._telescope_list[0].get_coordinates(
                crs_name=crs_name, coordinate_field="name"
            )
            table[_name_x] = pos_x
            table[_name_y] = pos_y
            table[_name_z] = pos_z
            if len(pos_t) > 0:
                table["telescope_axis_height"] = pos_t
            if len(tel_r) > 0:
                table["sphere_radius"] = tel_r
        except IndexError:
            pass

        if "telescope_name" in table.colnames:
            table.sort("telescope_name")
        if "asset_code" in table.colnames:
            table.sort(["asset_code", "sequence_number"])

        return table

    def export_one_telescope_as_json(
        self,
        crs_name,
        parameter_version=None,
        schema_version=None,
    ):
        """
        Return a list containing a single telescope in simtools-DB-style json.

        Parameters
        ----------
        crs_name: str
            Name of coordinate system to be used for export.
        schema_version: str
            Version of the schema.

        Returns
        -------
        dict
            Dictionary with array element information.
        """
        table = self.export_telescope_list_table(crs_name)
        if len(table) != 1:
            raise ValueError("Only one telescope can be exported to json")
        parameter_name = value = None
        if crs_name == "ground":
            parameter_name = "array_element_position_ground"
            value = [
                table["position_x"][0].value,
                table["position_y"][0].value,
                table["position_z"][0].value,
            ]
        elif crs_name == "utm":
            parameter_name = "array_element_position_utm"
            value = [
                table["utm_east"][0].value,
                table["utm_north"][0].value,
                table["altitude"][0].value,
            ]
        elif crs_name == "mercator":
            parameter_name = "array_element_position_mercator"
            value = [
                table["latitude"][0].value,
                table["longitude"][0].value,
                table["altitude"][0].value,
            ]

        return {
            "schema_version": schema.get_model_parameter_schema_version(schema_version),
            "parameter": parameter_name,
            "instrument": table["telescope_name"][0],
            "site": self.site,
            "parameter_version": parameter_version,
            "unique_id": None,
            "value": value,
            "unit": "m",
            "type": "float64",
            "file": False,
            "meta_parameter": False,
            "model_parameter_schema_version": "0.1.0",
        }

    def get_number_of_telescopes(self):
        """
        Return the number of telescopes in the list.

        Returns
        -------
        int
            Number of telescopes.
        """
        return len(self._telescope_list)

    def print_telescope_list(self, crs_name):
        """
        Print list of telescopes.

        Parameters
        ----------
        crs_name: str
            Name of coordinate system to be used for export.

        """
        for tel in self._telescope_list:
            tel.print_compact_format(
                crs_name=crs_name,
                print_header=(tel == self._telescope_list[0]),
                corsika_observation_level=(
                    self._corsika_observation_level if crs_name == "ground" else None
                ),
            )

    def convert_coordinates(self):
        """Perform all the possible conversions the coordinates of the tel positions."""
        self._logger.info("Converting telescope coordinates")

        crs_wgs84 = self.geo_coordinates.crs_wgs84()
        crs_local = self.geo_coordinates.crs_local(self._array_center)
        crs_utm = self.geo_coordinates.crs_utm(self._reference_position_dict.get("epsg_code", None))

        for tel in self._telescope_list:
            tel.convert_all(
                crs_local=crs_local,
                crs_wgs84=crs_wgs84,
                crs_utm=crs_utm,
            )

    def select_assets(self, asset_list=None):
        """
        Select a subsets of telescopes / assets from the layout.

        Parameters
        ----------
        asset_list: list
            List of assets to be selected (telescope names or types)

        Raises
        ------
        ValueError
            If the asset list is empty.

        """
        _n_telescopes = len(self._telescope_list)
        try:
            if len(asset_list) > 0:
                _telescope_list_from_name = [
                    tel for tel in self._telescope_list if tel.asset_code in asset_list
                ]
                _telescope_list_from_type = [
                    tel
                    for tel in self._telescope_list
                    if names.get_array_element_type_from_name(tel.asset_code) in asset_list
                ]
                self._telescope_list = list(
                    set(_telescope_list_from_name + _telescope_list_from_type)
                )
            self._logger.info(
                f"Selected {len(self._telescope_list)} telescopes (from originally {_n_telescopes})"
            )
        except TypeError:
            self._logger.info("No asset list provided, keeping all telescopes")



#!/usr/bin/python3
"""Plots for light emission (flasher/calibration) sim_telarray events."""

import logging
from pathlib import Path

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from ctapipe.calib import CameraCalibrator
from ctapipe.io import EventSource
from ctapipe.visualization import CameraDisplay
from scipy import signal as _signal

from simtools.data_model.metadata_collector import MetadataCollector
from simtools.visualization.plot_corsika_histograms import save_figs_to_pdf
from simtools.visualization.visualize import save_figure

_logger = logging.getLogger(__name__)

# Reusable literal constants (duplicated from visualize to avoid circular deps)
AXES_FRACTION = "axes fraction"
NO_R1_WAVEFORMS_MSG = "No R1 waveforms available in event"
TIME_NS_LABEL = "time [ns]"
R1_SAMPLES_LABEL = "R1 samples [d.c.]"

# Choices understood by the dispatcher used below
PLOT_CHOICES = {
    "event_image": "event_image",
    "time_traces": "time_traces",
    "waveform_matrix": "waveform_matrix",
    "step_traces": "step_traces",
    "integrated_signal_image": "integrated_signal_image",
    "integrated_pedestal_image": "integrated_pedestal_image",
    "peak_timing": "peak_timing",
    "all": "all",
}


def _get_event_source_and_r1_tel(filename, event_index=None, warn_context=None):
    """Return (source, event, first_r1_tel_id) or None if unavailable.

    Centralizes creation of EventSource, event selection, and first R1 tel-id lookup.

    When no event exists, logs a standard warning. When the event has no R1 tel data,
    logs either a contextual message ("Event has no R1 data for <context>") if
    warn_context is provided, or the generic "First event has no R1 telescope data".
    """
    source = EventSource(filename, max_events=None)
    event = _select_event_by_type(source)(event_index=event_index)
    if not event:
        _logger.warning("No event found in the file.")
        return None

    tel_ids = sorted(getattr(event.r1, "tel", {}).keys())
    if not tel_ids:
        if warn_context:
            _logger.warning("Event has no R1 data for %s", warn_context)
        else:
            _logger.warning("First event has no R1 telescope data")
        return None
    return source, event, int(tel_ids[0])


def _compute_integration_window(peak_idx, n_samp, half_width, mode, offset):
    """Return [a, b) window bounds for integration for signal/pedestal modes."""
    hw = int(half_width)
    win_len = 2 * hw + 1
    if mode == "signal":
        a = max(0, peak_idx - hw)
        b = min(n_samp, peak_idx + hw + 1)
        return a, b

    g = int(offset) if offset is not None else 16
    start = peak_idx + g
    if start + win_len <= n_samp:
        return start, start + win_len
    start = max(0, peak_idx - g - win_len)
    a = start
    b = min(n_samp, start + win_len)
    if a >= b:
        return 0, min(n_samp, win_len)
    return a, b


def _format_integrated_title(tel_label, et_name, half_width, mode, offset):
    win_len = 2 * int(half_width) + 1
    if mode == "signal":
        return f"{tel_label} integrated signal (win {win_len}) ({et_name})"
    g = int(offset) if offset is not None else 16
    return f"{tel_label} integrated pedestal (win {win_len}, offset {g}) ({et_name})"


def _select_event_by_type(source):
    """
    Build an event selector for a ctapipe EventSource.

    Parameters
    ----------
    source : ctapipe.io.EventSource
        Iterable event source.

    Returns
    -------
    callable
        A function ``select_event(event_index: int | None) -> Any`` that returns
        the first event (when ``event_index`` is None) or the event at the given
        index. Returns ``None`` if no event is available or the index is out of range.
    """

    def select_event(event_index=None):
        if event_index is None:
            for ev in source:
                return ev
        else:
            for idx, ev in enumerate(source):
                if idx == event_index:
                    return ev
        _logger.warning("No events available from source or event_index out of range")
        return None

    return select_event


def _time_axis_from_readout(readout, n_samp):
    """
    Compute time axis in nanoseconds from a camera readout.

    Parameters
    ----------
    readout : Any
        Camera readout providing ``sampling_rate`` as an astropy Quantity.
    n_samp : int
        Number of samples per trace.

    Returns
    -------
    numpy.ndarray
        Array of shape ``(n_samp,)`` with time in nanoseconds.
    """
    try:
        dt = (1 / readout.sampling_rate).to(u.ns).value
    except (AttributeError, TypeError):
        dt = 1.0
    if not np.isfinite(dt) or dt <= 0:
        dt = 1.0
    return np.arange(int(n_samp)) * float(dt)


def plot_simtel_event_image(filename, distance=None, event_index=None):
    """
    Read a sim_telarray file and plot the DL1 image for one event via ctapipe.

    Parameters
    ----------
    filename : str | pathlib.Path
        Path to the ``.simtel`` file.
    distance : astropy.units.Quantity | float | None, optional
        Distance to annotate in the plot. If a Quantity, interpreted in meters.
        If not provided, no unit conversion is attempted.
    event_index : int | None, optional
        Zero-based index of the event to plot. If None, the first event is used.

    Returns
    -------
    matplotlib.figure.Figure | None
        The created figure, or ``None`` if no suitable event/image is available.
    """
    prepared = _get_event_source_and_r1_tel(filename, event_index=event_index, warn_context=None)
    if prepared is None:
        return None
    source, event, tel_id = prepared

    calib = CameraCalibrator(subarray=source.subarray)
    calib(event)

    geometry = source.subarray.tel[tel_id].camera.geometry
    image = event.dl1.tel[tel_id].image

    fig, ax = plt.subplots(dpi=300)
    tel = source.subarray.tel[tel_id]
    tel_label = getattr(tel, "name", f"CT{tel_id}")
    title = f"{tel_label}, run {event.index.obs_id} event {event.index.event_id}"
    disp = CameraDisplay(geometry, image=image, norm="symlog", ax=ax)
    disp.cmap = "RdBu_r"
    disp.add_colorbar(fraction=0.02, pad=-0.1)
    disp.set_limits_percent(100)
    ax.set_title(title, pad=20)

    try:
        d_str = f"{distance.to(u.m)}"
    except (AttributeError, TypeError, ValueError):
        d_str = str(distance)

    ax.annotate(
        f"tel type: {source.subarray.tel[tel_id].type.name}\n"
        f"optics: {source.subarray.tel[tel_id].optics.name}\n"
        f"camera: {source.subarray.tel[tel_id].camera_name}\n"
        f"distance: {d_str}",
        xy=(0, 0),
        xytext=(0.1, 1),
        xycoords=AXES_FRACTION,
        va="top",
        size=7,
    )
    ax.annotate(
        f"dl1 image,\ntotal ADC counts: {np.round(np.sum(image))}\n",
        xy=(0, 0),
        xytext=(0.75, 1),
        xycoords=AXES_FRACTION,
        va="top",
        ha="left",
        size=7,
    )
    ax.set_axis_off()
    fig.tight_layout()
    return fig


def plot_simtel_time_traces(
    filename,
    tel_id=None,
    n_pixels=3,
    event_index=None,
):
    """
    Plot R1 time traces for a few pixels of one event.

    Parameters
    ----------
    filename : str | pathlib.Path
        Path to the ``.simtel`` file.
    tel_id : int | None, optional
        Telescope ID to use. If None, the first telescope with R1 data is chosen.
    n_pixels : int, optional
        Number of pixels with highest signal to plot. Default is 3.
    event_index : int | None, optional
        Zero-based index of the event to plot. If None, the first event is used.

    Returns
    -------
    matplotlib.figure.Figure | None
        The created figure, or ``None`` if R1 waveforms are unavailable.
    """
    prepared = _get_event_source_and_r1_tel(
        filename, event_index=event_index, warn_context="time traces plot"
    )
    if prepared is None:
        return None
    source, event, tel_id_default = prepared
    tel_id = tel_id or tel_id_default

    calib = CameraCalibrator(subarray=source.subarray)
    try:
        calib(event)
        image = event.dl1.tel[tel_id].image
    except (RuntimeError, ValueError, KeyError, AttributeError):
        image = None

    waveforms = getattr(event.r1.tel.get(tel_id, None), "waveform", None)
    if waveforms is None:
        _logger.warning(NO_R1_WAVEFORMS_MSG)
        return None

    w = np.asarray(waveforms)
    if w.ndim == 3:
        w = w[0]
    _, n_samp = w.shape

    if image is not None:
        pix_ids = np.argsort(image)[-n_pixels:][::-1]
    else:
        integrals = w.sum(axis=1)
        pix_ids = np.argsort(integrals)[-n_pixels:][::-1]

    readout = source.subarray.tel[tel_id].camera.readout
    t = _time_axis_from_readout(readout, n_samp)

    fig, ax = plt.subplots(dpi=300)
    for pid in pix_ids:
        ax.plot(t, w[pid], label=f"pix {int(pid)}", drawstyle="steps-mid")
    ax.set_xlabel(TIME_NS_LABEL)
    ax.set_ylabel(R1_SAMPLES_LABEL)
    et_name = getattr(getattr(event.trigger, "event_type", None), "name", "?")
    tel = source.subarray.tel[tel_id]
    tel_label = getattr(tel, "name", f"CT{tel_id}")
    ax.set_title(f"{tel_label} waveforms ({et_name})")
    ax.legend(loc="best", fontsize=7)
    fig.tight_layout()
    return fig


def plot_simtel_waveform_matrix(
    filename,
    tel_id=None,
    vmax=None,
    event_index=None,
    pixel_step=None,
):
    """
    Create a pseudocolor image of R1 waveforms (sample index vs. pixel id).

    Parameters
    ----------
    filename : str | pathlib.Path
        Path to the ``.simtel`` file.
    tel_id : int | None, optional
        Telescope ID to use. If None, the first telescope with R1 data is chosen.
    vmax : float | None, optional
        Upper limit for color normalization. If None, determined automatically.
    event_index : int | None, optional
        Zero-based index of the event to plot. If None, the first event is used.
    pixel_step : int | None, optional
        Step between plotted pixel ids (e.g., 1 plots all, 2 plots every second pixel).

    Returns
    -------
    matplotlib.figure.Figure | None
        The created figure, or ``None`` if R1 waveforms are unavailable.
    """
    prepared = _get_event_source_and_r1_tel(
        filename, event_index=event_index, warn_context="waveform plot"
    )
    if prepared is None:
        return None
    source, event, tel_id_default = prepared
    tel_id = tel_id or tel_id_default

    waveforms = getattr(event.r1.tel.get(tel_id, None), "waveform", None)
    if waveforms is None:
        _logger.warning(NO_R1_WAVEFORMS_MSG)
        return None

    w = np.asarray(waveforms)
    if w.ndim == 3:
        w = w[0]
    n_pix, n_samp = w.shape

    step = max(1, int(pixel_step)) if pixel_step is not None else 1
    pix_idx = np.arange(n_pix)[::step]
    w_sel = w[pix_idx]

    readout = source.subarray.tel[tel_id].camera.readout
    t = _time_axis_from_readout(readout, n_samp)

    fig, ax = plt.subplots(dpi=300)
    mesh = ax.pcolormesh(t, pix_idx, w_sel, shading="auto", vmax=vmax)
    cbar = fig.colorbar(mesh, ax=ax)
    cbar.set_label(R1_SAMPLES_LABEL)
    et_name = getattr(getattr(event.trigger, "event_type", None), "name", "?")
    tel = source.subarray.tel[tel_id]
    tel_label = getattr(tel, "name", f"CT{tel_id}")
    ax.set_title(f"{tel_label} waveform matrix ({et_name})")
    ax.set_xlabel(TIME_NS_LABEL)
    ax.set_ylabel("pixel id")
    fig.tight_layout()
    return fig


def plot_simtel_step_traces(
    filename,
    tel_id=None,
    pixel_step=100,
    max_pixels=None,
    event_index=None,
):
    """
    Plot step-style R1 traces for regularly sampled pixels (0, N, 2N, ...).

    Parameters
    ----------
    filename : str | pathlib.Path
        Path to the ``.simtel`` file.
    tel_id : int | None, optional
        Telescope ID to use. If None, the first telescope with R1 data is chosen.
    pixel_step : int, optional
        Interval between pixel indices to plot. Default is 100.
    max_pixels : int | None, optional
        Maximum number of pixels to plot. If None, plot all selected by ``pixel_step``.
    event_index : int | None, optional
        Zero-based index of the event to plot. If None, the first event is used.

    Returns
    -------
    matplotlib.figure.Figure | None
        The created figure, or ``None`` if R1 waveforms are unavailable.
    """
    prepared = _get_event_source_and_r1_tel(
        filename, event_index=event_index, warn_context="traces plot"
    )
    if prepared is None:
        return None
    source, event, tel_id_default = prepared
    tel_id = tel_id or tel_id_default

    waveforms = getattr(event.r1.tel.get(tel_id, None), "waveform", None)
    if waveforms is None:
        _logger.warning(NO_R1_WAVEFORMS_MSG)
        return None

    w = np.asarray(waveforms)
    if w.ndim == 3:
        w = w[0]
    n_pix, n_samp = w.shape

    readout = source.subarray.tel[tel_id].camera.readout
    t = _time_axis_from_readout(readout, n_samp)

    pix_ids = np.arange(0, n_pix, max(1, pixel_step))
    if max_pixels is not None:
        pix_ids = pix_ids[:max_pixels]

    fig, ax = plt.subplots(dpi=300)
    for pid in pix_ids:
        ax.plot(t, w[int(pid)], label=f"pix {int(pid)}", drawstyle="steps-mid")
    ax.set_xlabel(TIME_NS_LABEL)
    ax.set_ylabel(R1_SAMPLES_LABEL)
    et_name = getattr(getattr(event.trigger, "event_type", None), "name", "?")
    tel = source.subarray.tel[tel_id]
    tel_label = getattr(tel, "name", f"CT{tel_id}")
    ax.set_title(f"{tel_label} step traces ({et_name})")
    ax.legend(loc="best", fontsize=7, ncol=2)
    fig.tight_layout()
    return fig


def _detect_peaks(trace, peak_width, signal_mod):
    """
    Detect peak indices using CWT if available, otherwise ``find_peaks``.

    Parameters
    ----------
    trace : numpy.ndarray
        One-dimensional waveform samples for a single pixel.
    peak_width : int
        Characteristic peak width (samples) for CWT.
    signal_mod : module
        SciPy ``signal``-like module providing ``find_peaks_cwt``/``find_peaks``.

    Returns
    -------
    numpy.ndarray
        Array of integer indices of detected peaks (possibly empty).
    """
    peaks = []
    try:
        if hasattr(signal_mod, "find_peaks_cwt"):
            peaks = signal_mod.find_peaks_cwt(trace, widths=np.array([peak_width]))
        if not np.any(peaks):
            peaks, _ = signal_mod.find_peaks(trace, prominence=np.max(trace) * 0.1)
    except (ValueError, TypeError):
        peaks = []
    return np.asarray(peaks, dtype=int) if np.size(peaks) else np.array([], dtype=int)


def _collect_peak_samples(w, sum_threshold, peak_width, signal_mod):
    """
    Compute peak-sample indices per pixel from waveform matrix.

    Parameters
    ----------
    w : numpy.ndarray
        Waveform array of shape ``(n_pix, n_samples)`` (or ``(1, n_pix, n_samples)``).
    sum_threshold : float
        Minimum sum over samples for a pixel to be considered.
    peak_width : int
        Characteristic peak width (samples) for CWT.
    signal_mod : module
        SciPy ``signal``-like module providing peak finding routines.

    Returns
    -------
    tuple[numpy.ndarray | None, numpy.ndarray | None, int]
        ``(peak_samples, pix_ids, found_count)`` where ``peak_samples`` are the
        selected peak indices per considered pixel, ``pix_ids`` are the pixel
        indices that passed ``sum_threshold``, and ``found_count`` is the number
        of pixels with at least one detected peak. Returns ``(None, None, 0)`` if
        no pixels passed the threshold.
    """
    n_pix, _ = w.shape
    sums = w.sum(axis=1)
    has_signal = sums > float(sum_threshold)
    pix_ids = np.arange(n_pix)[has_signal]
    if pix_ids.size == 0:
        return None, None, 0

    peak_samples = []
    found_count = 0
    for pid in pix_ids:
        trace = w[int(pid)]
        pks = _detect_peaks(trace, peak_width, signal_mod)
        if pks.size:
            found_count += 1
            peak_idx = int(pks[np.argmax(trace[pks])])
        else:
            peak_idx = int(np.argmax(trace))
        peak_samples.append(peak_idx)

    return np.asarray(peak_samples), pix_ids, found_count


def _histogram_edges(n_samp, timing_bins):
    """
    Compute contiguous histogram bin edges for sample indices.

    Parameters
    ----------
    n_samp : int
        Number of samples per trace.
    timing_bins : int | None
        Number of histogram bins. If None, use unit-width bins.

    Returns
    -------
    numpy.ndarray
        Array of bin edges spanning the sample index range.
    """
    if timing_bins and timing_bins > 0:
        return np.linspace(-0.5, n_samp - 0.5, int(timing_bins) + 1)
    return np.arange(-0.5, n_samp + 0.5, 1.0)


def _draw_peak_hist(
    ax,
    peak_samples,
    edges,
    mean_sample,
    std_sample,
    tel_label,
    et_name,
    considered,
    found_count,
):
    """
    Draw a histogram of peak samples with overlays and annotations.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Target axes to draw into.

        Peak sample indices per pixel.
    edges : numpy.ndarray
        Histogram bin edges.
    mean_sample : float
        Mean of peak sample indices.
    std_sample : float
        Standard deviation of peak sample indices.
    tel_label : str
        Telescope label used in the title.
    et_name : str
        Event type name used in the title.
    considered : int
        Number of pixels considered (passed threshold).
    found_count : int
        Number of pixels with at least one detected peak.

    Returns
    -------
    None
    """
    counts, edges = np.histogram(peak_samples, bins=edges)
    ax.bar(edges[:-1], counts, width=np.diff(edges), color="#5B90DC", align="edge")
    ax.set_xlim(edges[0], edges[-1])
    ax.set_xlabel("peak sample")
    ax.set_ylabel("N pixels")
    ax.axvline(
        mean_sample,
        color="#D8153C",
        linestyle="--",
        label=f"mean={mean_sample:.2f}",
    )
    ax.axvspan(
        mean_sample - std_sample,
        mean_sample + std_sample,
        color="#D8153C",
        alpha=0.2,
        label=f"std={std_sample:.2f}",
    )
    ax.set_title(f"{tel_label} peak timing ({et_name})")
    ax.text(
        0.98,
        0.95,
        f"considered: {considered}\npeaks found: {found_count}",
        transform=ax.transAxes,
        ha="right",
        va="top",
        fontsize=7,
        bbox={
            "boxstyle": "round,pad=0.2",
            "facecolor": "white",
            "alpha": 0.6,
            "linewidth": 0.0,
        },
    )
    ax.legend(fontsize=7)


def plot_simtel_peak_timing(
    filename,
    tel_id=None,
    sum_threshold=10.0,
    peak_width=8,
    examples=3,
    timing_bins=None,
    return_stats=False,
    event_index=None,
):
    """
    Peak finding per pixel; report mean/std of peak sample and plot a histogram.

    Parameters
    ----------
    filename : str | pathlib.Path
        Path to the ``.simtel`` file.
    tel_id : int | None, optional
        Telescope ID to use. If None, the first telescope with R1 data is chosen.
    sum_threshold : float, optional
        Minimum sum over samples for a pixel to be considered. Default is 10.0.
    peak_width : int, optional
        Characteristic peak width (samples) for CWT. Default is 8.
    examples : int, optional
        Number of example pixel traces to overlay. Default is 3.
    timing_bins : int | None, optional
        Number of histogram bins. If None, use unit-width bins.
    return_stats : bool, optional
        If True, also return a statistics dictionary. Default is False.
    event_index : int | None, optional
        Zero-based index of the event to plot. If None, the first event is used.

    Returns
    -------
    matplotlib.figure.Figure | tuple[matplotlib.figure.Figure, dict] | None
        The created figure, or ``None`` if R1 waveforms are unavailable. If
        ``return_stats`` is True, a tuple ``(fig, stats)`` is returned, where
        ``stats`` has keys ``{"considered", "found", "mean", "std"}``.
    """
    prepared = _get_event_source_and_r1_tel(
        filename, event_index=event_index, warn_context="peak timing plot"
    )
    if prepared is None:
        return None
    source, event, tel_id_default = prepared
    tel_id = tel_id or tel_id_default

    waveforms = getattr(event.r1.tel.get(tel_id, None), "waveform", None)
    if waveforms is None:
        _logger.warning(NO_R1_WAVEFORMS_MSG)
        return None

    w = np.asarray(waveforms)
    if w.ndim == 3:
        w = w[0]
    _, n_samp = w.shape

    peak_samples, pix_ids, found_count = _collect_peak_samples(
        w, sum_threshold, peak_width, _signal
    )
    if peak_samples is None or pix_ids is None:
        _logger.warning("No pixels exceeded sum_threshold for peak timing")
        return None

    mean_sample = float(np.mean(peak_samples))
    std_sample = float(np.std(peak_samples))

    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 4), dpi=300)

    edges = _histogram_edges(n_samp, timing_bins)
    et_name = getattr(getattr(event.trigger, "event_type", None), "name", "?")
    tel = source.subarray.tel[tel_id]
    tel_label = getattr(tel, "name", f"CT{tel_id}")
    _draw_peak_hist(
        ax1,
        peak_samples,
        edges,
        mean_sample,
        std_sample,
        tel_label,
        et_name,
        pix_ids.size,
        found_count,
    )

    readout = source.subarray.tel[tel_id].camera.readout
    t = _time_axis_from_readout(readout, n_samp)

    ex_ids = pix_ids[: max(1, int(examples))]
    for pid in ex_ids:
        trace = w[int(pid)]
        pks = _detect_peaks(trace, peak_width, _signal)
        ax2.plot(t, trace, drawstyle="steps-mid", label=f"pix {int(pid)}")
        if pks.size:
            ax2.scatter(t[pks], trace[pks], s=10)
    ax2.set_xlabel(TIME_NS_LABEL)
    ax2.set_ylabel(R1_SAMPLES_LABEL)
    ax2.legend(fontsize=7)

    fig.tight_layout()

    if return_stats:
        stats = {
            "considered": int(pix_ids.size),
            "found": int(found_count),
            "mean": float(mean_sample),
            "std": float(std_sample),
        }
        return fig, stats
    return fig


def _prepare_waveforms_for_image(filename, tel_id, context_no_r1, event_index=None):
    """
    Fetch R1 waveforms for one event/telescope and return prepared arrays.

    Parameters
    ----------
    filename : str | pathlib.Path
        Path to the ``.simtel`` file.
    tel_id : int | None
        Telescope ID to use. If None, the first telescope with R1 data is chosen.
    context_no_r1 : str
        Short description used in warnings when no R1 data is available.
    event_index : int | None, optional
        Zero-based index of the event to use. If None, the first event is used.

    Returns
    -------
    tuple | None
        ``(w, n_pix, n_samp, source, event, tel_id)`` where ``w`` is a
        ``numpy.ndarray`` of shape ``(n_pix, n_samples)``, ``n_pix`` and
        ``n_samp`` are integers, and ``source``, ``event`` and ``tel_id`` are
        the ctapipe objects used. Returns ``None`` on failure.
    """
    prepared = _get_event_source_and_r1_tel(
        filename, event_index=event_index, warn_context=context_no_r1
    )
    if prepared is None:
        return None
    source, event, tel_id_default = prepared
    tel_id = tel_id or tel_id_default

    waveforms = getattr(event.r1.tel.get(tel_id, None), "waveform", None)
    if waveforms is None:
        _logger.warning(NO_R1_WAVEFORMS_MSG)
        return None

    w = np.asarray(waveforms)
    if w.ndim == 3:
        w = w[0]
    n_pix, n_samp = w.shape
    return w, n_pix, n_samp, source, event, tel_id


def plot_simtel_integrated_signal_image(
    filename,
    tel_id=None,
    half_width=8,
    event_index=None,
):
    """Plot camera image of integrated signal per pixel around the flasher peak."""
    return _plot_simtel_integrated_image(
        filename=filename,
        tel_id=tel_id,
        half_width=half_width,
        event_index=event_index,
        mode="signal",
    )


def plot_simtel_integrated_pedestal_image(
    filename,
    tel_id=None,
    half_width=8,
    offset=16,
    event_index=None,
):
    """Plot camera image of integrated pedestal per pixel away from the flasher peak."""
    return _plot_simtel_integrated_image(
        filename=filename,
        tel_id=tel_id,
        half_width=half_width,
        event_index=event_index,
        mode="pedestal",
        offset=offset,
    )


def _plot_simtel_integrated_image(
    filename,
    tel_id,
    half_width,
    event_index,
    mode,
    offset=None,
):
    """Shared implementation for integrated signal/pedestal images.

    mode: "signal" or "pedestal". For "pedestal", ``offset`` is used.
    """
    context = "integrated-signal image" if mode == "signal" else "integrated-pedestal image"
    prepared = _prepare_waveforms_for_image(filename, tel_id, context, event_index=event_index)
    if prepared is None:
        return None

    w, n_pix, n_samp, source, event, tel_id = prepared
    img = np.zeros(n_pix, dtype=float)

    for pid in range(n_pix):
        trace = w[pid]
        peak_idx = int(np.argmax(trace))
        a, b = _compute_integration_window(peak_idx, n_samp, half_width, mode, offset)
        img[pid] = float(np.sum(trace[a:b]))

    geometry = source.subarray.tel[tel_id].camera.geometry
    fig, ax = plt.subplots(dpi=300)
    disp = CameraDisplay(geometry, image=img, norm="lin", ax=ax)
    disp.cmap = "viridis" if mode == "signal" else "cividis"
    disp.add_colorbar(fraction=0.02, pad=-0.1)
    disp.set_limits_percent(100)

    et_name = getattr(getattr(event.trigger, "event_type", None), "name", "?")
    tel = source.subarray.tel[tel_id]
    tel_label = getattr(tel, "name", f"CT{tel_id}")
    ax.set_title(_format_integrated_title(tel_label, et_name, half_width, mode, offset), pad=20)
    ax.set_axis_off()
    fig.tight_layout()
    return fig


def _make_output_paths(ioh, base, input_file):
    """Return (out_dir, pdf_path) based on base name and input file."""
    out_dir = ioh.get_output_directory()
    pdf_path = ioh.get_output_file(f"{base}_{input_file.stem}" if base else input_file.stem)
    pdf_path = Path(f"{pdf_path}.pdf") if Path(pdf_path).suffix != ".pdf" else Path(pdf_path)
    return out_dir, pdf_path


def _call_peak_timing(
    filename,
    tel_id=None,
    sum_threshold=10.0,
    peak_width=8,
    examples=3,
    timing_bins=None,
    event_index=None,
):
    """Call plot_simtel_peak_timing while tolerating older signature.

    Returns a matplotlib Figure or None.
    """
    try:
        fig_stats = plot_simtel_peak_timing(
            filename,
            tel_id=tel_id,
            sum_threshold=sum_threshold,
            peak_width=peak_width,
            examples=examples,
            timing_bins=timing_bins,
            return_stats=True,
            event_index=event_index,
        )
        return fig_stats[0] if isinstance(fig_stats, tuple) else fig_stats
    except TypeError:
        return plot_simtel_peak_timing(
            filename,
            tel_id=tel_id,
            sum_threshold=sum_threshold,
            peak_width=peak_width,
            examples=examples,
            timing_bins=timing_bins,
            event_index=event_index,
        )


def _collect_figures_for_file(
    filename,
    plots,
    args,
    out_dir,
    base_stem,
    save_pngs,
    dpi,
):
    """Generate selected plots for a single sim_telarray file.

    Returns a list of figures. If ``save_pngs`` is True, also writes PNGs to ``out_dir``.
    """
    figures = []

    def add(fig, tag):
        if fig is not None:
            figures.append(fig)
            if save_pngs:
                base_path = out_dir / f"{base_stem}_{tag}"
                try:
                    save_figure(fig, base_path, figure_format=["png"], dpi=int(dpi))
                except Exception as ex:  # pylint:disable=broad-except
                    _logger.warning("Failed to save PNG %s: %s", base_path.with_suffix(".png"), ex)
        else:
            _logger.warning("Plot '%s' returned no figure for %s", tag, filename)

    plots_to_run = (
        [
            "event_image",
            "time_traces",
            "waveform_matrix",
            "step_traces",
            "integrated_signal_image",
            "integrated_pedestal_image",
            "peak_timing",
        ]
        if "all" in plots
        else list(plots)
    )

    dispatch = {
        "event_image": (
            plot_simtel_event_image,
            {"distance": None, "event_index": None},
        ),
        "time_traces": (
            plot_simtel_time_traces,
            {"tel_id": None, "n_pixels": 3, "event_index": None},
        ),
        "waveform_matrix": (
            plot_simtel_waveform_matrix,
            {"tel_id": None, "vmax": None, "event_index": None},
        ),
        "step_traces": (
            plot_simtel_step_traces,
            {"tel_id": None, "pixel_step": None, "max_pixels": None, "event_index": None},
        ),
        "integrated_signal_image": (
            plot_simtel_integrated_signal_image,
            {"tel_id": None, "half_width": 8, "event_index": None},
        ),
        "integrated_pedestal_image": (
            plot_simtel_integrated_pedestal_image,
            {"tel_id": None, "half_width": 8, "offset": 16, "event_index": None},
        ),
        "peak_timing": (
            _call_peak_timing,
            {
                "tel_id": None,
                "sum_threshold": 10.0,
                "peak_width": 8,
                "examples": 3,
                "timing_bins": None,
                "event_index": None,
            },
        ),
    }

    for plot_name in plots_to_run:
        entry = dispatch.get(plot_name)
        if entry is None:
            _logger.warning("Unknown plot selection '%s'", plot_name)
            continue
        func, defaults = entry
        kwargs = {k: args.get(k, v) for k, v in defaults.items()}
        fig = func(filename, **kwargs)  # type: ignore[misc]
        add(fig, plot_name)

    return figures


def generate_and_save_plots(
    simtel_files,
    plots,
    args,
    ioh,
):
    """Generate plots for files and save a multi-page PDF per input.

    Also writes metadata JSON next to the PDF.
    """
    for simtel in simtel_files:
        out_dir, pdf_path = _make_output_paths(ioh, args.get("output_file"), simtel)
        figures = _collect_figures_for_file(
            filename=simtel,
            plots=plots,
            args=args,
            out_dir=out_dir,
            base_stem=simtel.stem,
            save_pngs=bool(args.get("save_pngs", False)),
            dpi=int(args.get("dpi", 300)),
        )

        if not figures:
            _logger.warning("No figures produced for %s", simtel)
            continue

        try:
            save_figs_to_pdf(figures, pdf_path)
            _logger.info("Saved PDF: %s", pdf_path)
        except Exception as ex:  # pylint:disable=broad-except
            _logger.error("Failed to save PDF %s: %s", pdf_path, ex)

        try:
            MetadataCollector.dump(args, pdf_path, add_activity_name=True)
        except Exception as ex:  # pylint:disable=broad-except
            _logger.warning("Failed to write metadata for %s: %s", pdf_path, ex)



#!/usr/bin/python3

r"""
Derive CORSIKA configuration limits for energy, core distance, and viewcone radius.

This tool determines configuration limits based on triggered events from broad-range
simulations. It supports the derivation of the following CORSIKA configuration parameters:

- **ERANGE**: lower energy limit
- **CSCAT**: upper core distance
- **VIEWCONE**: viewcone radius

Broad-range simulations in this context are simulation sets generated with wide-ranging
definitions for above parameters.
Limits are computed based on a configurable maximum event loss fraction.
Results are provided as a table with the following columns:

+---------------------+-----------+--------+-----------------------------------------------+
| Field               | Data Type | Units  | Description                                   |
+=====================+===========+========+===============================================+
| primary_particle    | string    |        | Particle type (e.g., gamma, proton).          |
+---------------------+-----------+--------+-----------------------------------------------+
| array_name          | string    |        | Array name (custom or as defined in           |
|                     |           |        | 'array_layouts').                             |
+---------------------+-----------+--------+-----------------------------------------------+
| telescope_ids       | string    |        | Comma-separated list of telescope IDs         |
|                     |           |        | of this array.                                |
+---------------------+-----------+--------+-----------------------------------------------+
| zenith              | float64   | deg    | Direction of array pointing zenith.           |
+---------------------+-----------+--------+-----------------------------------------------+
| azimuth             | float64   | deg    | Direction of array pointing azimuth.          |
+---------------------+-----------+--------+-----------------------------------------------+
| nsb_level           | float64   |        | Night sky background level.                   |
+---------------------+-----------+--------+-----------------------------------------------+
| lower_energy_limit  | float64   | TeV    | Derived lower energy limit (**ERANGE**)       |
+---------------------+-----------+--------+-----------------------------------------------+
| upper_radius_limit  | float64   | m      | Derived upper core distance limit (**CSCAT**) |
+---------------------+-----------+--------+-----------------------------------------------+
| viewcone_radius     | float64   | deg    | Derived viewcone radius limit (**VIEWCONE**)  |
+---------------------+-----------+--------+-----------------------------------------------+

The input event data files are generated using the application simtools-generate-simtel-event-data
and are required for each point in the observational parameter space (e.g., array pointing
directions, level of night sky background, etc.).

Command line arguments
----------------------
event_data_file (str, required)
    Path to reduced event data file.
telescope_ids (str, optional)
    Custom array layout file containing telescope IDs.
loss_fraction (float, required)
    Maximum event-loss fraction for limit computation.
plot_histograms (bool, optional)
    Plot histograms of the event data.
output_file (str, optional)
    Path to the output file for the derived limits.

Example
-------

Derive limits for a list of array layouts (use 'all' to derive limits for all layouts):

.. code-block:: console

    simtools-production-derive-corsika-limits \\
        --event_data_file event_dat_file.hdf5 \\
        --array_layout_name alpha,beta \\
        --loss_fraction 1e-6 \\
        --plot_histograms \\
        --output_file corsika_simulation_limits_lookup.ecsv

Derive limits for a given file for custom defined array layouts:

.. code-block:: console

    simtools-production-derive-corsika-limits \\
        --event_data_file event_dat_file.hdf5 \\
        --telescope_ids path/to/telescope_configs.yaml \\
        --loss_fraction 1e-6 \\
        --plot_histograms \\
        --output_file corsika_simulation_limits_lookup.ecsv
"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.production_configuration.derive_corsika_limits import (
    generate_corsika_limits_grid,
)


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Derive limits for energy, radial distance, and viewcone.",
    )
    config.parser.add_argument(
        "--event_data_file",
        type=str,
        required=True,
        help="Event data file containing reduced event data.",
    )
    config.parser.add_argument(
        "--telescope_ids",
        type=str,
        required=False,
        help="Path to a file containing telescope configurations.",
    )
    config.parser.add_argument(
        "--loss_fraction",
        type=float,
        required=True,
        help="Maximum event-loss fraction for limit computation.",
    )
    config.parser.add_argument(
        "--plot_histograms",
        help="Plot histograms of the event data.",
        action="store_true",
        default=False,
    )
    return config.initialize(
        db_config=True,
        output=True,
        simulation_model=[
            "site",
            "model_version",
            "layout",
        ],
    )


def main():
    """Derive limits for energy, radial distance, and viewcone."""
    app_context = startup_application(_parse)

    generate_corsika_limits_grid(app_context.args, app_context.db_config)


if __name__ == "__main__":
    main()



"""Plot simtel event histograms filled with SimtelIOEventHistograms."""

import logging

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LogNorm

from simtools.simtel.simtel_io_event_histograms import SimtelIOEventHistograms

_logger = logging.getLogger(__name__)


def plot(histograms, output_path=None, limits=None, rebin_factor=2, array_name=None):
    """
    Plot simtel event histograms.

    Parameters
    ----------
    histograms: SimtelIOEventHistograms
        Instance containing the histograms to plot.
    output_path: Path or str, optional
        Directory to save plots. If None, plots will be displayed.
    limits: dict, optional
        Dictionary containing limits for plotting. Keys can include:
        - "upper_radius_limit": Upper limit for core distance
        - "lower_energy_limit": Lower limit for energy
        - "viewcone_radius": Radius for the viewcone
    rebin_factor: int, optional
        Factor by which to reduce the number of bins in 2D histograms for re-binned plots.
        Default is 2 (merge every 2 bins). Set to 0 or 1 to disable re-binning.
    array_name: str, optional
        Name of the telescope array configuration.
    """
    _logger.info(f"Plotting histograms written to {output_path}")

    plots = _generate_plot_configurations(histograms, limits)
    _execute_plotting_loop(plots, output_path, rebin_factor, array_name)


def _get_limits(name, limits):
    """
    Extract limits from the provided dictionary for plotting.

    Fine tuned to expected histograms to be plotted.
    """

    def _safe_value(limits, key):
        val = limits.get(key)
        return getattr(val, "value", None)

    mapping = {
        "energy": {"x": _safe_value(limits, "lower_energy_limit")},
        "core_distance": {"x": _safe_value(limits, "upper_radius_limit")},
        "angular_distance": {"x": _safe_value(limits, "viewcone_radius")},
        "core_vs_energy": {
            "x": _safe_value(limits, "upper_radius_limit"),
            "y": _safe_value(limits, "lower_energy_limit"),
        },
        "angular_distance_vs_energy": {
            "x": _safe_value(limits, "viewcone_radius"),
            "y": _safe_value(limits, "lower_energy_limit"),
        },
        "x_core_shower_vs_y_core_shower": {"r": _safe_value(limits, "upper_radius_limit")},
    }
    return mapping.get(name)


def _generate_plot_configurations(histograms, limits):
    """Generate plot configurations for all histogram types."""
    hist_1d_params = {"color": "tab:green", "edgecolor": "tab:green", "lw": 1}
    hist_2d_params = {"norm": "log", "cmap": "viridis", "show_contour": False}
    hist_2d_normalized_params = {"norm": "linear", "cmap": "viridis", "show_contour": True}
    plots = {}
    for name, hist in histograms.items():
        if hist["histogram"] is None:
            continue
        if hist["1d"]:
            plots[name] = _create_1d_plot_config(
                hist, name=name, plot_params=hist_1d_params, limits=limits
            )
        else:
            if "cumulative" in name or "efficiency" in name:
                plot_params = hist_2d_normalized_params
            else:
                plot_params = hist_2d_params

            plots[name] = _create_2d_plot_config(
                hist, name=name, plot_params=plot_params, limits=limits
            )
    return plots


def _get_axis_title(axis_titles, axis):
    """Return axis title for given axis."""
    if axis_titles is None:
        return None
    if axis == "x" and len(axis_titles) > 0:
        return axis_titles[0]
    if axis == "y" and len(axis_titles) > 1:
        return axis_titles[1]
    if axis == "z" and len(axis_titles) > 2:
        return axis_titles[2]
    return None


def _create_1d_plot_config(histogram, name, plot_params, limits):
    """Create a 1D plot configuration."""
    _logger.debug(f"Creating plot config for {name} with params: {plot_params}")
    return {
        "data": histogram["histogram"],
        "bins": histogram["bin_edges"],
        "plot_type": "histogram",
        "plot_params": plot_params,
        "labels": {
            "x": _get_axis_title(histogram.get("axis_titles"), "x"),
            "y": _get_axis_title(histogram.get("axis_titles"), "y"),
            "title": f"{histogram['title']}: {name.replace('_', ' ')}",
        },
        "scales": histogram["plot_scales"],
        "lines": _get_limits(name, limits) if limits else {},
        "filename": name,
    }


def _create_2d_plot_config(histogram, name, plot_params, limits):
    """Create a 2D plot configuration."""
    _logger.debug(f"Creating plot config for {name} with params: {plot_params}")
    return {
        "data": histogram["histogram"],
        "bins": [histogram["bin_edges"][0], histogram["bin_edges"][1]],
        "plot_type": "histogram2d",
        "plot_params": plot_params,
        "labels": {
            "x": _get_axis_title(histogram.get("axis_titles"), "x"),
            "y": _get_axis_title(histogram.get("axis_titles"), "y"),
            "title": f"{histogram['title']}: {name.replace('_', ' ')}",
        },
        "lines": _get_limits(name, limits) if limits else {},
        "scales": histogram["plot_scales"],
        "colorbar_label": _get_axis_title(histogram.get("axis_titles"), "z"),
        "filename": name,
    }


def _execute_plotting_loop(plots, output_path, rebin_factor, array_name):
    """Execute the main plotting loop for all plot configurations."""
    for plot_key, plot_args in plots.items():
        plot_filename = plot_args.pop("filename")

        if plot_args.get("data") is None:
            _logger.warning(f"Skipping plot {plot_key} - no data available")
            continue

        if array_name and plot_args.get("labels", {}).get("title"):
            plot_args["labels"]["title"] += f" ({array_name} array)"

        filename = _build_plot_filename(plot_filename, array_name)
        output_file = output_path / filename if output_path else None
        result = _create_plot(**plot_args, output_file=output_file)

        # Skip re-binned plot if main plot failed
        if result is None:
            continue

        if _should_create_rebinned_plot(rebin_factor, plot_args, plot_key):
            _create_rebinned_plot(plot_args, filename, output_path, rebin_factor)


def _build_plot_filename(base_filename, array_name=None):
    """
    Build the full plot filename with appropriate extensions.

    Parameters
    ----------
    base_filename : str
        The base filename without extension
    array_name : str, optional
        Name of the array to append to filename

    Returns
    -------
    str
        Complete filename with extension
    """
    return f"{base_filename}_{array_name}.png" if array_name else f"{base_filename}.png"


def _should_create_rebinned_plot(rebin_factor, plot_args, plot_key):
    """
    Check if a re-binned version of the plot should be created.

    Parameters
    ----------
    rebin_factor : int
        Factor by which to rebin the energy axis
    plot_args : dict
        Plot arguments
    plot_key : str
        Key identifying the plot type

    Returns
    -------
    bool
        True if a re-binned plot should be created, False otherwise
    """
    return (
        rebin_factor > 1
        and plot_args["plot_type"] == "histogram2d"
        and plot_key.endswith("_cumulative")
        and plot_args.get("plot_params", {}).get("norm") == "linear"
    )


def _create_rebinned_plot(plot_args, filename, output_path, rebin_factor):
    """
    Create a re-binned version of a 2D histogram plot.

    Parameters
    ----------
    plot_args : dict
        Plot arguments for the original plot
    filename : str
        Filename of the original plot
    output_path : Path or None
        Path to save the plot to, or None
    rebin_factor : int
        Factor by which to rebin the energy axis
    """
    data = plot_args["data"]
    bins = plot_args["bins"]

    rebinned_data, rebinned_x_bins, rebinned_y_bins = SimtelIOEventHistograms.rebin_2d_histogram(
        data, bins[0], bins[1], rebin_factor
    )

    rebinned_plot_args = plot_args.copy()
    rebinned_plot_args["data"] = rebinned_data
    rebinned_plot_args["bins"] = [rebinned_x_bins, rebinned_y_bins]

    if rebinned_plot_args.get("labels", {}).get("title"):
        rebinned_plot_args["labels"]["title"] += f" (Energy rebinned {rebin_factor}x)"

    rebinned_filename = f"{filename.replace('.png', '')}_rebinned.png"
    rebinned_output_file = output_path / rebinned_filename if output_path else None
    _create_plot(**rebinned_plot_args, output_file=rebinned_output_file)


def _create_plot(
    data,
    bins=None,
    plot_type="histogram",
    plot_params=None,
    labels=None,
    scales=None,
    colorbar_label=None,
    output_file=None,
    lines=None,
):
    """Create and save a plot with the given parameters."""
    plot_params = plot_params or {}
    labels = labels or {}
    scales = scales or {}
    lines = lines or {}

    if not _has_data(data):
        return None

    fig, ax = plt.subplots(figsize=(8, 6))
    _plot_data(ax, data, bins, plot_type, plot_params, colorbar_label)
    _add_lines(ax, lines)
    ax.set(
        xlabel=labels.get("x", ""),
        ylabel=labels.get("y", ""),
        title=labels.get("title", ""),
        xscale=scales.get("x", "linear"),
        yscale=scales.get("y", "linear"),
    )
    if output_file:
        _logger.info(f"Saving plot to {output_file}")
        fig.savefig(output_file, dpi=300, bbox_inches="tight")
        plt.close(fig)
    else:
        plt.tight_layout()
        plt.show()

    return fig


def _has_data(data):
    """Check that the data for plotting is not None or empty."""
    if data is None or (isinstance(data, np.ndarray) and data.size == 0):
        _logger.warning("No data available for plotting")
        return False
    return True


def _plot_data(ax, data, bins, plot_type, plot_params, colorbar_label):
    """Plot the data on the given axes."""
    if plot_type == "histogram":
        ax.bar(bins[:-1], data, width=np.diff(bins), **plot_params)
    elif plot_type == "histogram2d":
        pcm = _create_2d_histogram_plot(data, bins, plot_params)
        plt.colorbar(pcm, label=colorbar_label)


def _add_lines(ax, lines):
    """Add reference lines to the plot."""
    if lines.get("x") is not None:
        ax.axvline(lines["x"], color="r", linestyle="--", linewidth=0.5)
    if lines.get("y") is not None:
        ax.axhline(lines["y"], color="r", linestyle="--", linewidth=0.5)
    if lines.get("r") is not None:
        ax.add_artist(
            plt.Circle((0, 0), lines["r"], color="r", fill=False, linestyle="--", linewidth=0.5)
        )


def _create_2d_histogram_plot(data, bins, plot_params):
    """
    Create a 2D histogram plot with the given parameters.

    Parameters
    ----------
    data : np.ndarray
        2D histogram data
    bins : tuple of np.ndarray
        Bin edges for x and y axes
    plot_params : dict
        Plot parameters including norm, cmap, and show_contour

    Returns
    -------
    matplotlib.collections.QuadMesh
        The created pcolormesh object for colorbar attachment
    """
    if plot_params.get("norm") == "linear":
        pcm = plt.pcolormesh(
            bins[0],
            bins[1],
            data.T,
            vmin=0,
            vmax=1,
            cmap=plot_params.get("cmap", "viridis"),
        )
        # Add contour line at value=1.0 for normalized histograms
        if plot_params.get("show_contour", True):
            x_centers = (bins[0][1:] + bins[0][:-1]) / 2
            y_centers = (bins[1][1:] + bins[1][:-1]) / 2
            x_mesh, y_mesh = np.meshgrid(x_centers, y_centers)
            plt.contour(
                x_mesh,
                y_mesh,
                data.T,
                levels=[0.999999],  # very close to 1 for floating point precision
                colors=["tab:red"],
                linestyles=["--"],
                linewidths=[0.5],
            )
    else:
        # Handle empty or invalid data for logarithmic scaling
        data_max = data.max()
        if data_max <= 0:
            _logger.warning("No positive data found for logarithmic scaling, using linear scale")
            pcm = plt.pcolormesh(
                bins[0], bins[1], data.T, vmin=0, vmax=max(1, data_max), cmap="viridis"
            )
        else:
            # Ensure vmin is less than vmax for LogNorm
            vmin = max(1, data[data > 0].min()) if np.any(data > 0) else 1
            vmax = max(vmin + 1, data_max)
            pcm = plt.pcolormesh(
                bins[0], bins[1], data.T, norm=LogNorm(vmin=vmin, vmax=vmax), cmap="viridis"
            )

    return pcm



#!/usr/bin/python3
"""Definition of site model."""

import logging
from pathlib import Path

import numpy as np
from astropy import units as u

from simtools.model.model_parameter import ModelParameter


class SiteModel(ModelParameter):
    """
    Representation of an observatory site model.

    The site model includes (among others):

    - Reference point coordinates
    - Array elements
    - Geomagnetic field parameters
    - Atmospheric parameters
    - NSB parameters

    Parameters
    ----------
    site: str
        Site name (e.g., South or North).
    db_config: dict
        Database configuration.
    model_version: str or list
        Model version or list of model versions (in which case only the first one is used).
    label: str, optional
        Instance label.
    overwrite_model_parameters: str, optional
        File name to overwrite model parameters from DB with provided values.
    ignore_software_version: bool, optional
        If True, ignore software version checks for deprecated parameters.
    """

    def __init__(
        self,
        site,
        db_config,
        model_version,
        label=None,
        overwrite_model_parameters=None,
        ignore_software_version=False,
    ):
        """Initialize SiteModel."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init SiteModel for site %s", site)
        super().__init__(
            site=site,
            db_config=db_config,
            model_version=model_version,
            label=label,
            collection="sites",
            overwrite_model_parameters=overwrite_model_parameters,
            ignore_software_version=ignore_software_version,
        )

    def get_reference_point(self):
        """
        Get reference point coordinates.

        Ground coordinates are calculated relative to this point.

        Returns
        -------
        dict
            Reference point coordinates.
        """
        return {
            "center_altitude": self.get_parameter_value_with_unit("reference_point_altitude"),
            "center_northing": self.get_parameter_value_with_unit("reference_point_utm_north"),
            "center_easting": self.get_parameter_value_with_unit("reference_point_utm_east"),
            "epsg_code": self.get_parameter_value("epsg_code"),
        }

    def get_corsika_site_parameters(self, config_file_style=False, model_directory=None):
        """
        Get site-related CORSIKA parameters.

        Parameters are returned with units wherever possible ('config_file_style=False')
        or in CORSIKA-expected coordinates.

        Parameters
        ----------
        config_file_style: bool
            Return using CORSIKA config file syntax
        model_directory: Path, optional
            Model directory to use for file paths

        Returns
        -------
        dict
            Site-related CORSIKA parameters.
        """
        if config_file_style:
            model_directory = model_directory or Path()
            return {
                "OBSLEV": [
                    self.get_parameter_value_with_unit("corsika_observation_level").to_value("cm")
                ],
                # We always use a custom profile by filename, so this has to be set to 99
                "ATMOSPHERE": [99, "Y"],
                "IACT ATMOFILE": [
                    model_directory / self.get_parameter_value("atmospheric_profile")
                ],
                "MAGNET": [
                    self.get_parameter_value("geomag_horizontal"),
                    self.get_parameter_value("geomag_vertical"),
                ],
                "ARRANG": [self.get_parameter_value("geomag_rotation")],
            }

        return {
            "corsika_observation_level": self.get_parameter_value_with_unit(
                "corsika_observation_level"
            ),
            "geomag_horizontal": self.get_parameter_value_with_unit("geomag_horizontal"),
            "geomag_vertical": self.get_parameter_value_with_unit("geomag_vertical"),
            "geomag_rotation": self.get_parameter_value_with_unit("geomag_rotation"),
        }

    def get_array_elements_for_layout(self, layout_name):
        """
        Return list of array elements for a given array layout.

        Parameters
        ----------
        layout_name: str
            Name of the array layout

        Returns
        -------
        list
            List of array elements
        """
        layouts = self.get_parameter_value("array_layouts")
        for layout in layouts:
            if layout["name"] == layout_name.lower():
                return layout["elements"]
        raise ValueError(f"Array layout '{layout_name}' not found in '{self.site}' site model.")

    def get_array_elements_of_type(self, array_element_type):
        """
        Get all array elements of a given type.

        Parameters
        ----------
        array_element_type : str
            Type of the array element (e.g. LSTN, MSTS)

        Returns
        -------
        dict
            Dict with array elements.
        """
        return self.db.get_array_elements_of_type(
            array_element_type=array_element_type,
            model_version=self.model_version,
            collection="telescopes",
        )

    def get_list_of_array_layouts(self):
        """
        Get list of available array layouts.

        Returns
        -------
        list
            List of available array layouts
        """
        return [layout["name"] for layout in self.get_parameter_value("array_layouts")]

    def export_atmospheric_transmission_file(self, model_directory):
        """
        Export atmospheric transmission file from database to the given directory.

        Parameters
        ----------
        model_directory: Path
            Model directory to export the file to.
        """
        self.db.export_model_files(
            parameters={
                "atmospheric_transmission_file": {
                    "value": self.get_parameter_value("atmospheric_profile"),
                    "file": True,
                }
            },
            dest=model_directory,
        )

    def get_nsb_integrated_flux(self, wavelength_min=300 * u.nm, wavelength_max=650 * u.nm):
        """
        Get the integrated flux for the NSB (Night Sky Background) model.

        Returns
        -------
        float
            Integrated flux value.
        """
        table = self.db.get_ecsv_file_as_astropy_table(
            file_name=self.get_parameter_value("nsb_spectrum")
        )
        table.sort("wavelength")
        wl = table["wavelength"].quantity.to(u.nm)
        rate = table["differential photon rate"].quantity.to(1 / (u.nm * u.cm**2 * u.ns * u.sr))
        mask = (wl >= wavelength_min) & (wl <= wavelength_max)
        integral_cm2 = np.trapezoid(rate[mask], wl[mask])
        self._logger.debug(
            f"NSB integral between {wavelength_min} and {wavelength_max}: {integral_cm2}"
        )
        return integral_cm2.value



#!/usr/bin/python3
"""
Generate sim_telarray configuration files for a given array.

The applications generates the sim_telarray configuration files for a given array, site,
and model_version using the model parameters stored in the database.

Command line arguments
----------------------
site : str
    Site name (e.g., North, South).
array_layout_name : str
    Name of the layout array (e.g., test_layout, alpha, 4mst, etc.).
array_element_list : list
    List of array elements (e.g., telescopes) to plot (e.g., ``LSTN-01 LSTN-02 MSTN``).

Example
-------
North - 5.0.0:

.. code-block:: console

    simtools-generate-array-config --site North --array_layout_name alpha --model_version 5.0.0

The output is saved in simtools-output/test/model.
"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.model.array_model import ArrayModel


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Generate sim_telarray configuration files for a given array.",
    )
    return config.initialize(db_config=True, simulation_model=["site", "layout", "model_version"])


def main():
    """Generate sim_telarray configuration files for a given array."""
    app_context = startup_application(_parse)

    array_model = ArrayModel(
        label=app_context.args["label"],
        model_version=app_context.args["model_version"],
        db_config=app_context.db_config,
        site=app_context.args.get("site"),
        layout_name=app_context.args.get("array_layout_name"),
        array_elements=app_context.args.get("array_elements"),
    )
    array_model.print_telescope_list()
    array_model.export_all_simtel_config_files()


if __name__ == "__main__":
    main()



#!/usr/bin/python3
"""Read metadata from sim_telarray files."""

import logging
import re
from functools import cache

from eventio import EventIOFile
from eventio.iact import InputCard
from eventio.simtel import HistoryMeta

from simtools.utils import names

_logger = logging.getLogger(__name__)


@cache
def read_sim_telarray_metadata(file, encoding="utf8"):
    """
    Return global and per-telescope metadata from sim_telarray file.

    Cached to avoid re-reading the file multiple times.

    Parameters
    ----------
    file: str
        Path to the sim_telarray file.
    encoding: str
        Encoding to use for decoding metadata. Default is 'utf8'.

    Returns
    -------
    global_meta: dict
        Global metadata dictionary.
    telescope_meta: dict
        Dictionary of telescope metadata, keyed by telescope ID.
    """
    global_meta = None
    telescope_meta = {}

    with EventIOFile(file) as f:
        found_meta = False
        for o in f:
            if isinstance(o, HistoryMeta):
                found_meta = True
            else:
                if found_meta:
                    break
                continue

            meta = _decode_dictionary(o.parse(), encoding=encoding)
            if o.header.id == -1:
                global_meta = meta
            else:
                telescope_meta[o.header.id] = meta

    def clean_meta(meta):
        """Clean metadata dictionary."""
        meta = {k.lower().lstrip("*"): v for k, v in meta.items()}
        return {k: v.strip() if isinstance(v, str) else v for k, v in meta.items()}

    # keys to lower case and strip leading '*', trailing spaces
    try:
        return clean_meta(global_meta), {
            tel_id: clean_meta(meta) for tel_id, meta in telescope_meta.items()
        }
    except AttributeError as e:
        raise AttributeError(f"Error reading metadata from file {file}: {e}") from e


def _decode_dictionary(meta, encoding="utf8"):
    """Decode metadata dictionary."""

    def safe_decode(byte_str, encoding, errors="ignore"):
        return byte_str.decode(encoding, errors=errors)

    try:
        return {k.decode(encoding, errors="ignore"): v.decode(encoding) for k, v in meta.items()}
    except UnicodeDecodeError as e:
        _logger.warning(
            f"Unable to decode metadata with encoding {encoding}: {e}. "
            "Falling back to 'utf-8' with errors='ignore'."
        )
        return {safe_decode(k, encoding): safe_decode(v, encoding) for k, v in meta.items()}


def get_sim_telarray_telescope_id(telescope_name, file):
    """
    Return the telescope ID for a given telescope name in a sim_telarray file.

    Translates e.g. 'LSTN-01' to the corresponding telescope ID.

    Parameters
    ----------
    telescope_name: str
        Name of the telescope.
    file: str
        Path to the sim_telarray file.

    Returns
    -------
    int, None
        Telescope ID. Returns None if not found.
    """
    _, telescope_meta = read_sim_telarray_metadata(file)
    telescope_name_to_sim_telarray_id = {}
    for tel_id in telescope_meta.keys():
        _optics_name = telescope_meta[tel_id].get("optics_config_variant", None)
        _camera_name = telescope_meta[tel_id].get("camera_config_variant", None)
        if _optics_name == _camera_name and _optics_name == telescope_name:
            telescope_name_to_sim_telarray_id[telescope_name] = tel_id

    return telescope_name_to_sim_telarray_id.get(telescope_name, None)


def get_sim_telarray_telescope_id_to_telescope_name_mapping(file):
    """
    Return a mapping of telescope IDs to telescope names from a sim_telarray file.

    Parameters
    ----------
    file: str
        Path to the sim_telarray file.

    Returns
    -------
    dict
        Dictionary mapping telescope IDs to telescope names.
    """
    _, telescope_meta = read_sim_telarray_metadata(file)
    telescope_map = {}
    for i, (tel_id, meta) in enumerate(telescope_meta.items()):
        try:
            telescope_name = names.validate_array_element_name(
                meta.get("optics_config_variant", f"Unknown-{tel_id}")
            )
        except ValueError:
            telescope_name = _guess_telescope_name_for_legacy_files(i, file)
        if telescope_name is not None:
            telescope_map[tel_id] = telescope_name

    return telescope_map


def _guess_telescope_name_for_legacy_files(tel_counter, file):
    """
    Guess telescope names for legacy prod6 sim_telarray files with incomplete metadata.

    Parameters
    ----------
    tel_counter: int
        Telescope counter, used to index into the telescope list.
    file: str, Path
        Path to the sim_telarray file.

    Returns
    -------
    str, None
        Guessed telescope name or None if not found.
    """
    telescope_list = _get_telescope_list_from_input_card(file)
    try:
        return names.validate_array_element_name(telescope_list[tel_counter])
    except (IndexError, ValueError):
        pass
    return None


@cache
def _get_telescope_list_from_input_card(file):
    r"""
    Return telescope list from CORSIKA input card.

    Note hardwired regex pattern with telescope naming convention.
    This function is intended for legacy files generated for prod6,
    where metadata is incomplete.

    Expected format in input card:

    .. code-block:: console
        TELESCOPE    -70.91E2     -52.35E2 45.00E2  12.50E2  # (ID=1)  LSTN   01   2B5\n

    Parameters
    ----------
    file: str, Path
        Path to the sim_telarray file.

    Returns
    -------
    list
        List of telescope names as found in CORSIKA input card.
    """
    with EventIOFile(file) as f:
        for o in f:
            if isinstance(o, InputCard):
                input_card = o.parse().decode("utf-8")
                regex = (
                    r"TELESCOPE\s+[-\d.E]+\s+[-\d.E]+\s+[-\d.E]+\s+[-\d.E]+\s+"
                    r"# \(ID=\d+\)\s+(LST[N|S]|MST[N|S]|S[S|C]TS)\s+([^\s]+)"
                )
                return [f"{m[0]}-{m[1]}" for m in re.findall(regex, input_card)]
    return []



#!/usr/bin/python3

r"""
    Derive CTAO array layouts definitions using CTAO common identifiers repository.

    Retrieves the CTAO array layouts definitions from the CTAO common identifiers repository
    and synchronizes them with the model parameter definition in 'array_layouts' for all
    CTAO sites.

    Requires access to the CTAO common identifiers repository. Future versions of the
    common identifiers might be stored in a CTAO technical database.

    Command line arguments
    ----------------------
    site (str)
        CTAO site (North or South).
    parameter_version (str)
        Model parameter version.
    repository_url (str)
        URL or path of the CTAO common identifiers repository.
    repository_branch (str)
        Repository branch to use for CTAO common identifiers.
    updated_parameter_version (str)
        Updated parameter version.

    Example
    -------

    Derive the CTAO array layouts definitions for the North site using the CTAO common identifiers
    repository. Merge the retrieved array layouts with the model parameter definition in
    'array_layouts' (in this example for site North and parameter version 2.0.0). Write the
    updated array layouts and metadata using the parameter version indicated by the
    'updated_parameter_version' argument.

    .. code-block:: console

        simtools-derive-ctao-array-layouts --site North \
            --repository_url "https://gitlab.cta-observatory.org/" \
            "cta-computing/common/identifiers/-/raw/" \
            --repository_branch main \
            --site North --parameter_version 2.0.0 \
            --updated_parameter_version 3.0.0
"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.db import db_handler
from simtools.layout.array_layout_utils import (
    merge_array_layouts,
    retrieve_ctao_array_layouts,
    write_array_layouts,
)


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Derive CTAO array layouts from CTAO common identifiers repository.",
    )
    config.parser.add_argument(
        "--repository_url",
        help="URL or path of the CTAO common identifiers repository.",
        type=str,
        default="https://gitlab.cta-observatory.org/cta-computing/common/identifiers/-/raw/",
    )
    config.parser.add_argument(
        "--repository_branch",
        help="Repository branch to use for CTAO common identifiers.",
        type=str,
        default="main",
        required=False,
    )
    config.parser.add_argument(
        "--updated_parameter_version",
        help="Updated parameter version.",
        type=str,
        required=False,
    )
    return config.initialize(
        db_config=True, output=True, simulation_model=["site", "parameter_version", "model_version"]
    )


def main():
    """Derive CTAO array layouts from CTAO common identifiers repository."""
    app_context = startup_application(_parse)

    ctao_array_layouts = retrieve_ctao_array_layouts(
        site=app_context.args["site"],
        repository_url=app_context.args["repository_url"],
        branch_name=app_context.args["repository_branch"],
    )

    db = db_handler.DatabaseHandler(db_config=app_context.db_config)
    db_array_layouts = db.get_model_parameter(
        parameter="array_layouts",
        site=app_context.args["site"],
        array_element_name=None,
        parameter_version=app_context.args.get("parameter_version"),
        model_version=app_context.args.get("model_version"),
    )
    db_array_layouts["array_layouts"].pop("_id", None)
    db_array_layouts["array_layouts"].pop("entry_date", None)
    app_context.logger.info(f"Layouts from model parameter database: {db_array_layouts}")

    write_array_layouts(
        array_layouts=merge_array_layouts(db_array_layouts["array_layouts"], ctao_array_layouts),
        args_dict=app_context.args,
        db_config=app_context.db_config,
    )


if __name__ == "__main__":
    main()



"""A collection of functions related to geometrical transformations."""

import logging
import math

import astropy.units as u
import numpy as np
from astropy.units import UnitsError

_logger = logging.getLogger(__name__)


def convert_2d_to_radial_distr(hist_2d, xaxis, yaxis, bins=50, max_dist=1000):
    """
    Convert a 2d histogram of positions, e.g. photon positions on the ground, to a 1D distribution.

    Parameters
    ----------
    hist_2d: numpy.ndarray
        The histogram counts.
    xaxis: numpy.array
        The values of the x axis (histogram bin edges) on the ground.
    yaxis: numpy.array
        The values of the y axis (histogram bin edges) on the ground.
    bins: float
        Number of bins in distance.
    max_dist: float
       Maximum distance to consider in the 1D histogram, usually in meters.

    Returns
    -------
    np.array
        The values of the 1D histogram with size = int(max_dist/bin_size).
    np.array
        The bin edges of the 1D histogram with size = int(max_dist/bin_size) + 1.

    """
    # Check if the histogram will make sense
    bins_step = 2 * max_dist / bins  # in the 2d array, the positive and negative direction count.
    for axis in [xaxis, yaxis]:
        if (bins_step < np.diff(axis)).any():
            msg = (
                f"The histogram with number of bins {bins} and maximum distance of {max_dist} "
                f"resulted in a bin size smaller than the original array. Please adjust those "
                f"parameters to increase the bin size and avoid nan in the histogram values."
            )
            _logger.warning(msg)
            break

    grid_2d_x, grid_2d_y = np.meshgrid(xaxis[:-1], yaxis[:-1])  # [:-1], since xaxis and yaxis are
    # the hist bin_edges (n + 1).
    # radial_distance_map maps the distance to the center from each element in a square matrix.
    radial_distance_map = np.sqrt(grid_2d_x**2 + grid_2d_y**2)
    # The sorting and unravel_index give us the two indices for the position of the sorted element
    # in the original 2d matrix
    sorted_indices = np.unravel_index(
        np.argsort(radial_distance_map, axis=None), np.shape(radial_distance_map)
    )
    x_indices_sorted, y_indices_sorted = sorted_indices[0], sorted_indices[1]

    # We construct a 1D array with the histogram counts sorted according to the distance to the
    # center.
    hist_sorted = np.array(
        [hist_2d[i_x, i_y] for i_x, i_y in zip(x_indices_sorted, y_indices_sorted)]
    )
    distance_sorted = np.sort(radial_distance_map, axis=None)

    # For larger distances, we have more elements in a slice 'dr' in radius, hence, we need to
    # account for it using weights below.

    weights, radial_bin_edges = np.histogram(distance_sorted, bins=bins, range=(0, max_dist))
    histogram_1d = np.empty_like(weights, dtype=float)

    for i_radial, _ in enumerate(radial_bin_edges[:-1]):
        # Here we sum all the events within a radial interval 'dr' and then divide by the number of
        # bins that fit this interval.
        indices_to_sum = (distance_sorted >= radial_bin_edges[i_radial]) * (
            distance_sorted < radial_bin_edges[i_radial + 1]
        )
        if weights[i_radial] != 0:
            histogram_1d[i_radial] = np.sum(hist_sorted[indices_to_sum]) / weights[i_radial]
        else:
            histogram_1d[i_radial] = 0
    return histogram_1d, radial_bin_edges


@u.quantity_input(rotation_angle_phi=u.rad, rotation_angle_theta=u.rad)
def rotate(x, y, rotation_around_z_axis, rotation_around_y_axis=0):
    """
    Rotate the x and y coordinates of the telescopes.

    The two rotations are:

    - rotation_angle_around_z_axis gives the rotation on the observation plane (x, y)
    - rotation_angle_around_y_axis allows to rotate the observation plane in space.

    The function returns the rotated x and y values in the same unit given.
    The direction of rotation of the elements in the plane is counterclockwise, i.e.,
    the rotation of the coordinate system is clockwise.

    Parameters
    ----------
    x: numpy.array or list
        x positions of the entries (e.g. telescopes), usually in meters.
    y: numpy.array or list
        y positions of the entries (e.g. telescopes), usually in meters.
    rotation_angle_around_z_axis: astropy.units.rad
        Angle to rotate the array in the observation plane (around z axis) in radians.
    rotation_angle_around_y_axis: astropy.units.rad
        Angle to rotate the observation plane around the y axis in radians.

    Returns
    -------
    2-tuple of list
        x and y positions of the rotated entry (e.g. telescopes) positions.

    Raises
    ------
    TypeError:
        If type of x and y parameters are not valid.
    RuntimeError:
        If the length of x and y are different.
    UnitsError:
        If the unit of x and y are different.
    """
    allowed_types = (list, np.ndarray, u.Quantity, float, int)
    if not all(isinstance(variable, (allowed_types)) for variable in [x, y]):
        raise TypeError("x and y types are not valid! Cannot perform transformation.")

    if not isinstance(x, list | np.ndarray):
        x = [x]
    if not isinstance(y, list | np.ndarray):
        y = [y]

    if (
        np.sum(
            np.array([isinstance(x, type_now) for type_now in allowed_types[:-2]])
            * np.array([isinstance(y, type_now) for type_now in allowed_types[:-2]])
        )
        == 0
    ):
        raise TypeError("x and y are not from the same type! Cannot perform transformation.")

    if len(x) != len(y):
        raise RuntimeError(
            "Cannot perform coordinate transformation when x and y have different lengths."
        )
    if all(isinstance(variable, (u.Quantity)) for variable in [x, y]):
        if not isinstance(x[0].unit, type(y[0].unit)):
            raise UnitsError(
                "Cannot perform coordinate transformation when x and y have different units."
            )

    x_trans = np.cos(rotation_around_y_axis) * (
        x * np.cos(rotation_around_z_axis) - y * np.sin(rotation_around_z_axis)
    )
    y_trans = x * np.sin(rotation_around_z_axis) + y * np.cos(rotation_around_z_axis)

    return x_trans, y_trans


def calculate_circular_mean(angles):
    """
    Calculate circular mean of angles in radians.

    Parameters
    ----------
    angles: numpy.array
        Array of angles in radians.

    Returns
    -------
    float
        Circular mean of the angles.
    """
    sin_sum = np.sum(np.sin(angles))
    cos_sum = np.sum(np.cos(angles))
    return np.arctan2(sin_sum, cos_sum)


@u.quantity_input(angle_max=u.rad, angle_min=u.rad)
def solid_angle(angle_max, angle_min=0 * u.rad):
    """
    Calculate the solid angle subtended by a given range of angles.

    Parameters
    ----------
    angle_max: astropy.units.Quantity
        The maximum angle for which to calculate the solid angle.
    angle_min: astropy.units.Quantity
        The minimum angle for which to calculate the solid angle (default is 0 rad).

    Returns
    -------
    astropy.units.Quantity
        The solid angle subtended by the given range of angles (in steradians).
    """
    return 2 * np.pi * (np.cos(angle_min.to("rad")) - np.cos(angle_max.to("rad"))) * u.sr


def transform_ground_to_shower_coordinates(x_ground, y_ground, z_ground, azimuth, altitude):
    """
    Transform ground to shower coordinates.

    Assume ground to be of type 'North-West-Up' (NWU) coordinates.

    Parameters
    ----------
    x_ground: numpy.array
        Ground x coordinate.
    y_ground: numpy.array
        Ground y coordinate.
    z_ground: numpy.array
        Ground z coordinate.
    azimuth: numpy.array
        Azimuth angle of the shower (in radians).
    altitude: numpy.array
        Altitude angle of the shower (in radians).

    Returns
    -------
    tuple
        Transformed shower coordinates (x', y', z').
    """
    x, y, z, az, alt = np.broadcast_arrays(x_ground, y_ground, z_ground, azimuth, altitude)

    ca, sa = np.cos(az), np.sin(az)
    cz, sz = np.sin(alt), np.cos(alt)

    x_s = ca * cz * x - sa * y + ca * sz * z
    y_s = sa * cz * x + ca * y + sa * sz * z
    z_s = -sz * x + cz * z

    return x_s, y_s, z_s


def fiducial_radius_from_shape(width, shape):
    """
    Calculate minimum radius including different geometrical shapes.

    Assumes definition of shapes as in 'camera_body_shape' model parameter:

    - circle: shape = 0, width is diameter
    - hexagon: shape = 1 or 3, width is flat-to-flat distance
    - square: shape = 2, width is side length

    Parameters
    ----------
    width : float
        Characteristic width
    shape : int
        Geometrical shape parameter

    Returns
    -------
    float
        Minimum fiducial radius
    """
    if shape == 0:
        return width / 2.0
    if shape == 2:
        return width / math.sqrt(2.0)
    if shape in (1, 3):
        return width / math.sqrt(3.0)
    raise ValueError(
        f"Unknown shape value {shape}. Valid values are: 0 (circle), 1 or 3 (hexagon), 2 (square)."
    )
